# -*- coding: utf-8 -*-
"""SPHARA transform

This module provides a class to perform the SPHARA transform. The
class is derived from :class:`spharapy.spharabasis.SpharaBasis`. It
provides methodes the SPHARA anaylsis and synthesis of spatially
irregularly sampled data.

"""

import numpy as np
from scipy import linalg
from spharapy.spharabasis import SpharaBasis

__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.13"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


class SpharaTransform(SpharaBasis):
    """SPHARA transform class

    This class is used to perform the SPHARA forward (analysis) and
    inverse (synthesis) transformation.

    Parameters
    ----------
    triangsamples : trimesh object
        A trimesh object from the package spharapy in which the triangulation
        of the spatial arrangement of the sampling points is stored. The SPHARA
        basic functions are determined for this triangulation of the sample
        points.
    mode : {'unit', 'inv_euclidean', 'fem'}, optional
        The discretisation method used to estimate the Laplace-Beltrami
        operator. Using the option 'unit' all edges of
        the mesh are weighted by unit weighting function. The option
        'inv_euclidean' results in edge weights corresponding to the
        inverse Euclidean distance of the edge lengths. The option
        'fem' uses a FEM discretisation. the default weighting
        function is 'fem'.

    """

    def __init__(self, triangsamples=None, mode='fem'):
        SpharaBasis.__init__(self, triangsamples, mode)
        self._basis = None
        self._frequencies = None
        self._massmatrix = None

    def analysis(self, data):
        r"""Perform the SPHARA transform (analysis)

        This method performs the SPHARA transform (analysis) of data
        defined at spatially distributed sampling points described by
        a triangular mesh. The forward transformation is performed by
        matrix multiplication of the data matrix and the matrix with
        SPHARA basis functions :math:`\tilde{X} = X \cdot S`, with the
        SPHARA basis :math:`S`, the data matrix :math:`X` and the
        SPHARA coefficients matix :math:`\tilde{X}`. In the forward
        transformation using SPHARA basic functions determined by
        discretization with FEM approach, the modified scalar product
        including the mass matrix is used :math:`\tilde{X} = X \cdot B
        \cdot S`, with the mass matrix :math:`B`.

        Parameters
        ----------
        data : array, shape(m, n_points)
            A matrix with data to be transformed (analyzed) by
            SPHARA. The number of vertices of the triangular mesh is
            n_points. The order of the spatial sample points must
            correspond to that in the vertex list used to determine
            the SPHARA basis functions.

        Returns
        -------
        coefficients : array, shape (m, n_points)
            A matrix containing the SPHARA coefficients. The coefficients
            are sorted column by column with increasing spatial frequency,
            starting with DC in the first column.

        Examples
        --------

        Import the necessary packages

        >>> import numpy as np
        >>> from spharapy import trimesh as tm
        >>> from spharapy import spharatransform as st
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> st_fem_simple = st.SpharaTransform(testtrimesh, mode='fem')
        >>> data = np.concatenate([[[0., 0., 0.], [1., 1., 1.]],
        ...                        np.transpose(st_fem_simple.basis()[0])])
        >>> data
        array([[ 0.        ,  0.        ,  0.        ],
               [ 1.        ,  1.        ,  1.        ],
               [ 0.53452248,  0.53452248,  0.53452248],
               [-0.49487166, -0.98974332,  1.48461498],
               [ 1.42857143, -1.14285714, -0.28571429]])
        >>> coef_fem_simple = st_fem_simple.analysis(data)
        >>> coef_fem_simple
        array([[  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  1.87082869e+00,   1.09883582e-16,  -4.18977022e-16],
               [  1.00000000e+00,  -2.75573800e-16,  -8.86630311e-18],
               [ -1.14766454e-16,   1.00000000e+00,   2.30648330e-16],
               [  6.52367763e-17,   1.68383874e-16,   1.00000000e+00]])

        """

        # lazy evaluation, compute the basis at the first request and store
        # it until the triangular mesh or the discretization method is changed
        if self._basis is None or self._frequencies is None:
            self.basis()

        data = np.asarray(data)

        # Does the number of spatial samples of the data correspond to
        # that of the basis functions?
        if self._basis.shape[0] != data.shape[1]:
            raise ValueError('Dimension mismatch')

        # compute the SPHARA coefficients
        if self._mode == 'fem':
            coefficients = np.matmul(np.matmul(data, self._massmatrix),
                                     self._basis)
        else:
            coefficients = np.matmul(data, self._basis)

        return coefficients

    def synthesis(self, coefficients):
        r"""Perform the inverse SPHARA transform (synthesis)

        This method performs the inverse SPHARA transform (synthesis)
        for data defined at spatially distributed sampling points
        described by a triangular mesh. The forward transformation is
        performed by matrix multiplication of the data matrix and the
        matrix with SPHARA basis functions :math:`\tilde{X} = X \cdot S`,
        with the SPHARA basis :math:`S`, the data matrix :math:`X` and the
        SPHARA coefficients matix :math:`\tilde{X}`. In the forward
        transformation using SPHARA basic functions determined by
        discretization with FEM approach, the modified scalar product
        including the mass matrix is used
        :math:`\tilde{X} = X \cdot B \cdot S`, with the mass matrix
        :math:`B`.

        Parameters
        ----------
        coefficients : array, shape (m, n_points)
            A matrix containing the SPHARA coefficients. The coefficients
            are sorted column by column with increasing spatial frequency,
            starting with DC in the first column.

        Returns
        -------
        data : array, shape(m, n_points)
            A matrix with data to be forward transformed (analyzed) by
            SPHARA. The number of vertices of the triangular mesh is
            n_points. The order of the spatial sample points must correspond
            to that in the vertex list used to determine the SPHARA basis
            functions.

        Examples
        --------

        >>> import numpy as np
        >>> from spharapy import trimesh as tm
        >>> from spharapy import spharatransform as st
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> st_fem_simple = st.SpharaTransform(testtrimesh, mode='fem')
        >>> data = np.concatenate([[[0., 0., 0.], [1., 1., 1.]],
        ...                        np.transpose(st_fem_simple.basis()[0])])
        >>> data
        array([[ 0.        ,  0.        ,  0.        ],
               [ 1.        ,  1.        ,  1.        ],
               [ 0.53452248,  0.53452248,  0.53452248],
               [-0.49487166, -0.98974332,  1.48461498],
               [ 1.42857143, -1.14285714, -0.28571429]])
        >>> coef_fem_simple = st_fem_simple.analysis(data)
        >>> coef_fem_simple
        array([[  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  1.87082869e+00,   1.09883582e-16,  -4.18977022e-16],
               [  1.00000000e+00,  -2.75573800e-16,  -8.86630311e-18],
               [ -1.14766454e-16,   1.00000000e+00,   2.30648330e-16],
               [  6.52367763e-17,   1.68383874e-16,   1.00000000e+00]])
        >>> recon_fem_simple = st_fem_simple.synthesis(coef_fem_simple)
        >>> recon_fem_simple
        array([[ 0.        ,  0.        ,  0.        ],
               [ 1.        ,  1.        ,  1.        ],
               [ 0.53452248,  0.53452248,  0.53452248],
               [-0.49487166, -0.98974332,  1.48461498],
               [ 1.42857143, -1.14285714, -0.28571429]])

        """

        # lazy evaluation, compute the basis at the first request and store
        # it until the triangular mesh or the discretization method is changed
        if self._basis is None or self._frequencies is None:
            self.basis()

        coefficients = np.asarray(coefficients)

        # Does the number of SPHARA coefficients correspond to that of
        # the basis functions?
        if self._basis.shape[0] != coefficients.shape[1]:
            raise ValueError('Dimension mismatch')

        # compute the data from SPHARA coefficients
        data = np.matmul(coefficients, np.transpose(self._basis))

        return data



# -*- coding: utf-8 -*-
"""Triangular mesh data

This module provides a class for storing triangular meshes. Attributes
of the triangular mesh can be determined. In addition, methodes are available
to derive further information from the triangular grid.

"""

import numpy as np
import copy

__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.15"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


def area_triangle(vertex1, vertex2, vertex3):
    """Estimate the area of a triangle given by three vertices

    The area of the triangle given by three vertices is calculated by the half
    cross product formula.

    Parameters
    ----------
    vertex1 : array, shape (1, 3)
    vertex2 : array, shape (1, 3)
    vertex3 : array, shape (1, 3)


    Returns
    -------
    trianglearea : float
        Area of the triangle given by the three vertices.


    Examples
    --------

    >>> from spharapy import trimesh as tm
    >>> tm.area_triangle([1, 0, 0], [0, 1, 0], [0, 0, 1])
    0.8660254037844386

    """
    vertex1 = np.asarray(vertex1)
    vertex2 = np.asarray(vertex2)
    vertex3 = np.asarray(vertex3)

    trianglearea = (0.5 *
                    np.linalg.norm(np.cross(vertex2 - vertex1,
                                            vertex3 - vertex1)))
    return trianglearea


def side_lens_triangle(vertex1, vertex2, vertex3):
    """Estimate the three side length of a triangle given by three vertices

    Parameters
    ----------
    vertex1 : array, shape (1, 3)
    vertex2 : array, shape (1, 3)
    vertex3 : array, shape (1, 3)


    Returns
    -------
    side_lens : array, shape (1, 3)
        Side lengths of the triangle given by the three vertices.


    Examples
    --------

    >>> from spharapy import trimesh as tm
    >>> tm.side_lens_triangle([1, 0, 0], [0, 1, 0], [0, 0, 1])
    array([1.41421356, 1.41421356, 1.41421356])

    """
    vertex1 = np.asarray(vertex1)
    vertex2 = np.asarray(vertex2)
    vertex3 = np.asarray(vertex3)

    side_lens = np.array([np.linalg.norm(vertex2 - vertex1),
                          np.linalg.norm(vertex3 - vertex2),
                          np.linalg.norm(vertex1 - vertex3)])

    return side_lens


def angles_triangle(vertex1, vertex2, vertex3):
    """Estimate the three internal angles of a triangle given by three vertices

    Parameters
    ----------
    vertex1 : array, shape (1, 3)
    vertex2 : array, shape (1, 3)
    vertex3 : array, shape (1, 3)


    Returns
    -------
    angles : array, shape (1, 3)
        Internal angles of the triangle given by the three vertices.


    Examples
    --------

    >>> from spharapy import trimesh as tm
    >>> tm.angles_triangle([1, 0, 0], [0, 1, 0], [0, 0, 1])
    array([1.04719755, 1.04719755, 1.04719755])


    """
    vertex1 = np.asarray(vertex1)
    vertex2 = np.asarray(vertex2)
    vertex3 = np.asarray(vertex3)

    # compute unit vectors in direction of the triangle sides
    v1 = vertex2 - vertex1
    v1 = v1 / np.linalg.norm(v1)
    v2 = vertex3 - vertex2
    v2 = v2 / np.linalg.norm(v2)
    v3 = vertex1 - vertex3
    v3 = v3 / np.linalg.norm(v3)

    # compute the internal angles of the triangle
    angles = np.array([np.arccos(np.clip(np.dot(v1, -v3), -1.0, 1.0)),
                       np.arccos(np.clip(np.dot(v2, -v1), -1.0, 1.0)),
                       np.arccos(np.clip(np.dot(v3, -v2), -1.0, 1.0))])

    return angles


class TriMesh(object):
    """Triangular mesh class

    This class can be used to store data to define a triangular mesh and it
    provides atributes and methodes to derive further information about the
    triangular mesh.

    Parameters
    ----------
    trilist: array, shape (n_triangles, 3)
        List of triangles, each row of the array contains the edges of
        a triangle. The edges of the triangles are defined by the
        indices to the list of vertices. The index of the first vertex
        is 0. The number of triangles is n_triangles.
    vertlist: array, shape (n_points, 3)
        List of coordinates x, y, z which describes the positions of the
        vertices.

    Attributes
    ----------
    trilist: array, shape (n_triangles, 3)
        List of triangles of the mesh.
    vertlist: array, shape (n_points, 3)
        List of coordinates of the vertices
    """

    def __init__(self, trilist, vertlist):
        self.trilist = np.asarray(trilist)
        self.vertlist = np.asarray(vertlist)

    @property
    def trilist(self):
        """Get or set the list of triangles.

        Setting the list of triangles will simultaneously check if the
        triangle list is in the correct format.
        """

        return self._trilist

    @trilist.setter
    def trilist(self, trilist):
        if trilist.ndim != 2:
            raise ValueError('Triangle list has to be 2D!')
        elif trilist.shape[1] != 3:
            raise ValueError('Each entry of the triangle list has to consist '
                             'of three elements!')
        # pylint: disable=W0201
        self._trilist = np.asarray(trilist)

    @property
    def vertlist(self):
        """Get or set the list of vertices.

        Setting the list of triangles will simultaneously check if the
        vertice list is in the correct format.
        """

        return self._vertlist

    @vertlist.setter
    def vertlist(self, vertlist):
        if vertlist.ndim != 2:
            raise ValueError('Vertex list has to be 2D!')
        elif vertlist.shape[1] != 3:
            raise ValueError('Each entry of the vertex list has to consist '
                             'of three elements!')
        # pylint: disable=W0201
        self._vertlist = np.asarray(vertlist)

    def weightmatrix(self, mode='inv_euclidean'):
        """Compute a weight matrix for a triangular mesh

        The method creates a weighting matrix for the edges of a triangular
        mesh using different weighting function.

        Parameters
        ----------
        mode : {'unit', 'inv_euclidean', 'half_cotangent'}, optional
            The parameter `mode` specifies the method for determining
            the edge weights. Using the option 'unit' all edges of the
            mesh are weighted by unit weighting function, the result
            is an adjacency matrix. The option 'inv_euclidean' results
            in edge weights corresponding to the inverse Euclidean
            distance of the edge lengths. The option 'half_cotangent'
            uses the half of the cotangent of the two angles opposed
            to an edge as weighting function. the default weighting
            function is 'inv_euclidean'.

        Returns
        -------
        weightmatrix : array, shape (n_points, n_points)
            Symmetric matrix, which contains the weight of the edges
            between adjacent vertices. The number of vertices of the
            triangular mesh is n_points.

        Examples
        --------

        >>> from spharapy import trimesh as tm
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> testtrimesh.weightmatrix(mode='inv_euclidean')
        array([[ 0.        ,  0.4472136 ,  0.31622777],
               [ 0.4472136 ,  0.        ,  0.2773501 ],
               [ 0.31622777,  0.2773501 ,  0.        ]])

        """

        # plausibility test of option 'mode'
        if mode not in ('unit', 'inv_euclidean', 'half_cotangent'):
            raise ValueError("Unrecognized mode '%s'" % mode)

        # get the largest index from from triangle list
        maxindex = self.trilist.max()

        # fill the weight matrix with zeros, the size is (maxindex + 1)^2
        weightmatrix = np.zeros((maxindex + 1, maxindex + 1), dtype=float)

        if mode == 'unit':
            # iterate over triangle list an build weight matrix
            for x in self.trilist:

                weightmatrix[x[0], x[1]] = 1
                weightmatrix[x[1], x[0]] = 1

                weightmatrix[x[0], x[2]] = 1
                weightmatrix[x[2], x[0]] = 1

                weightmatrix[x[1], x[2]] = 1
                weightmatrix[x[2], x[1]] = 1

        elif mode == 'inv_euclidean':
            # iterate over triangle list an build weight matrix
            for x in self.trilist:
                # compute the three vectors of the triangle
                vec10 = self.vertlist[x[1]] - self.vertlist[x[0]]
                vec20 = self.vertlist[x[2]] - self.vertlist[x[0]]
                vec21 = self.vertlist[x[2]] - self.vertlist[x[1]]

                # fill in the weights in the weight matrix
                weightmatrix[x[0], x[1]] = (1 / np.linalg.norm(vec10))
                weightmatrix[x[1], x[0]] = (1 / np.linalg.norm(vec10))
                weightmatrix[x[0], x[2]] = (1 / np.linalg.norm(vec20))
                weightmatrix[x[2], x[0]] = (1 / np.linalg.norm(vec20))
                weightmatrix[x[2], x[1]] = (1 / np.linalg.norm(vec21))
                weightmatrix[x[1], x[2]] = (1 / np.linalg.norm(vec21))

        else:
            # iterate over triangle list an build weight matrix
            for x in self.trilist:
                # compute the directional vectors at the 1st vertex
                vec1 = self.vertlist[x[1]] - self.vertlist[x[0]]
                vec2 = self.vertlist[x[2]] - self.vertlist[x[0]]

                # compute the weight of the edge 0.5 * cot
                tempweight = 0.5 * (1.0 /
                                    np.tan(np.arccos(np.dot(vec1, vec2) /
                                                     (np.linalg.norm(vec1) *
                                                      np.linalg.norm(vec2)))))

                weightmatrix[x[1], x[2]] += tempweight
                weightmatrix[x[2], x[1]] += tempweight

                # compute the directional vectors at the 2nd vertex
                vec1 = self.vertlist[x[0]] - self.vertlist[x[1]]
                vec2 = self.vertlist[x[2]] - self.vertlist[x[1]]

                # compute the weight of the edge 0.5 * cot
                tempweight = 0.5 * (1.0 /
                                    np.tan(np.arccos(np.dot(vec1, vec2) /
                                                     (np.linalg.norm(vec1) *
                                                      np.linalg.norm(vec2)))))

                weightmatrix[x[0], x[2]] += tempweight
                weightmatrix[x[2], x[0]] += tempweight

                # compute the directional vectors at the 3rd vertex
                vec1 = self.vertlist[x[0]] - self.vertlist[x[2]]
                vec2 = self.vertlist[x[1]] - self.vertlist[x[2]]

                # compute the weight of the edge 0.5 * cot
                tempweight = 0.5 * (1.0 /
                                    np.tan(np.arccos(np.dot(vec1, vec2) /
                                                     (np.linalg.norm(vec1) *
                                                      np.linalg.norm(vec2)))))

                weightmatrix[x[0], x[1]] += tempweight
                weightmatrix[x[1], x[0]] += tempweight

        # return the weight matrix
        return weightmatrix

    def laplacianmatrix(self, mode='inv_euclidean'):
        """Compute a laplacian matrix for a triangular mesh

        The method creates a laplacian matrix for a triangular
        mesh using different weighting function.

        Parameters
        ----------
        mode : {'unit', 'inv_euclidean', 'half_cotangent'}, optional
            The methods for determining the edge weights. Using the option
            'unit' all edges of the mesh are weighted by unit weighting
            function, the result is an adjacency matrix. The option
            'inv_euclidean' results in edge weights corresponding to the
            inverse Euclidean distance of the edge lengths. The option
            'half_cotangent' uses the half of the cotangent of the two angles
            opposed to an edge as weighting function. the default weighting
            function is 'inv_euclidean'.

        Returns
        -------
        laplacianmatrix : array, shape (n_points, n_points)
            Matrix, which contains the discrete laplace operator for data
            defined at the vertices of a triangular mesh. The number of
            vertices of the triangular mesh is n_points.

        Examples
        --------

        >>> from spharapy import trimesh as tm
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> testtrimesh.laplacianmatrix(mode='inv_euclidean')
        array([[ 0.76344136, -0.4472136 , -0.31622777],
               [-0.4472136 ,  0.72456369, -0.2773501 ],
               [-0.31622777, -0.2773501 ,  0.59357786]])

        """

        # plausibility test of option 'mode'
        if mode not in ('unit', 'inv_euclidean', 'half_cotangent'):
            raise ValueError("Unrecognized mode '%s'" % mode)

        # determine the weight matrix with
        weightmatrix = self.weightmatrix(mode=mode)

        # compute the laplacian matrix
        laplacianmatrix = np.diag(weightmatrix.sum(axis=0)) - weightmatrix

        # return the laplacian matrix
        return laplacianmatrix

    def massmatrix(self, mode='normal'):
        """Mass matrix of a triangular mesh

        The method determines a mass matrix of a triangular mesh.

        Parameters
        ----------

        mode : {'normal', 'lumped'}, optional
            The `mode` parameter can be used to select whether a normal mass
            matrix or a lumped mass matrix is to be determined.

        Returns
        -------
        massmatrix : array, shape (n_points, n_points)
            Symmetric matrix, which contains the mass values for each edge and
            vertex for the FEM approch. The number of vertices of the
            triangular mesh is n_points.

        Examples
        --------

        >>> from spharapy import trimesh as tm
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> testtrimesh.massmatrix()
        array([[ 0.58333333,  0.29166667,  0.29166667],
               [ 0.29166667,  0.58333333,  0.29166667],
               [ 0.29166667,  0.29166667,  0.58333333]])

        References
        ----------
        :cite:`vallet07,dyer07,zhang07`

        """

        # plausibility test of option 'mode'
        if mode not in ('normal', 'lumped'):
            raise ValueError("Unrecognized mode '%s'" % mode)

        # get the largest index from from triangle list
        maxindex = self.trilist.max()

        # fill the weight matrix with zeros, the size is (maxindex + 1)^2
        massmatrix = np.zeros((maxindex + 1, maxindex + 1), dtype=float)

        if mode == 'lumped':
            # iterate over triangle list an build mass matrix
            for x in self.trilist:
                # compute the area of the triangle
                temparea = area_triangle(self.vertlist[x[0]],
                                         self.vertlist[x[1]],
                                         self.vertlist[x[2]])

                # add to every matrix element belonging to the vertex v(i) of
                # the triangle a 3rd of the triangle area
                massmatrix[x[0], x[0]] += temparea / 3
                massmatrix[x[1], x[1]] += temparea / 3
                massmatrix[x[2], x[2]] += temparea / 3
        else:
            # iterate over triangle list an build mass matrix
            for x in self.trilist:
                # compute the area of the triangle
                temparea = area_triangle(self.vertlist[x[0]],
                                         self.vertlist[x[1]],
                                         self.vertlist[x[2]])

                # add to every matrix element belonging to the edge e(i,j) of
                # the triangle a twelfth of the triangle area
                massmatrix[x[0], x[1]] += temparea / 12
                massmatrix[x[1], x[0]] = massmatrix[x[0], x[1]]
                massmatrix[x[0], x[2]] += temparea / 12
                massmatrix[x[2], x[0]] = massmatrix[x[0], x[2]]
                massmatrix[x[1], x[2]] += temparea / 12
                massmatrix[x[2], x[1]] = massmatrix[x[1], x[2]]

                # add to every matrix element belonging to the vertex v(i) of
                # the triangle a sixth of the triangle area
                massmatrix[x[0], x[0]] += temparea / 6
                massmatrix[x[1], x[1]] += temparea / 6
                massmatrix[x[2], x[2]] += temparea / 6

        # return the mass matrix
        return massmatrix

    def stiffnessmatrix(self):
        """Stiffness matrix of a triangular mesh

        The method determines a stiffness matrix of a triangular mesh.

        Returns
        -------
        stiffmatrix : array, shape (n_points, n_points)
            Symmetric matrix, which contains the stiffness values for each edge
            and vertex for the FEM approch. The number of vertices of the
            triangular mesh is n_points.

        Examples
        --------

        >>> from spharapy import trimesh as tm
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> testtrimesh.stiffnessmatrix()
        array([[-0.92857143,  0.64285714,  0.28571429],
               [ 0.64285714, -0.71428571,  0.07142857],
               [ 0.28571429,  0.07142857, -0.35714286]])

        References
        ----------
        :cite:`vallet07`

        """

        # get the largest index from from triangle list
        maxindex = self.trilist.max()

        # fill the weight matrix with zeros, the size is (maxindex + 1)^2
        stiffmatrix = np.zeros((maxindex + 1, maxindex + 1), dtype=float)

        # compute the cot weight matrix
        weightmatrix = self.weightmatrix(mode='half_cotangent')

        # compute and return the stiffness matrix
        stiffmatrix = -np.diag(weightmatrix.sum(axis=0)) + weightmatrix

        return stiffmatrix

    def one_ring_neighborhood(self, vertex_index=0):
        """The 1 ring neighborhood of a vertex

        The method determines all adjacent vertices of a vertex that
        is given by its index, the so called 1 ring neighborhood.

        Parameters
        ----------
        vertex_index : integer
            Index of the vertex for which the adjacent vertices are to
            be determined. The index must be in the range 0 to number of
            vertices - 1.

        Returns
        -------
        one_ring_neighborhood : array, shape (1, n)
            Array of indexes on vertices adjacent to a given vertex.
        """
        # plausibility test of vertex_index
        # is the index of type integer
        if not isinstance(vertex_index, (int, np.integer)):
            raise TypeError("""The parameter vertex_index has to be
            int.""")
        # is the index within the interval (0, maxindex_vertlist)
        if (vertex_index < 0) or (vertex_index > self.vertlist.shape[0]):
            raise IndexError("""The vertex_index is out of range.""")

        # get the adjacent triangles to a vertex
        adjacent_tri = \
            self.trilist[(self.trilist == vertex_index).any(axis=1)]

        # geth the indices of vertices of the triangles
        one_ring_neighborhood = np.unique(adjacent_tri)

        # remove the center vertex
        one_ring_neighborhood = \
            one_ring_neighborhood[one_ring_neighborhood != vertex_index]

        return one_ring_neighborhood

    def adjacent_tri(self, vertex_index=0):
        """All triangles with the given vertex

        The method determined all triangles of the triangular mesh that
        contain the given vertex.

        Parameters
        ----------
        vertex_index : integer
            Index of the vertex for which the adjacent vertices are to
            be determined. The index must be in the range 0 to number of
            vertices - 1.

        Returns
        -------
        tri_with_vertex : array, shape (3, n)
            List of triangles containing the given vertex.
        """
        # plausibility test of vertex_index
        # is the index of type integer
        if not isinstance(vertex_index, (int, np.integer)):
            raise TypeError("""The parameter vertex_index has to be
            int.""")
        # is the index within the interval (0, maxindex_vertlist)
        if (vertex_index < 0) or (vertex_index > self.vertlist.shape[0]):
            raise IndexError("""The vertex_index is out of range.""")

        # get the adjacent triangles to a vertex
        adjacent_tri = \
            self.trilist[(self.trilist == vertex_index).any(axis=1)]

        return adjacent_tri

    def is_edge(self, vertex1_index, vertex2_index):
        """Are 2 vertices connected by an edge

        The method determines whether two vertices are connected by an
        edge in the triangle mesh and if so, whether it is an internal
        edge or a boundary edge.

        Parameters
        ----------
        vertex1_index, vertex2_index : integer
            Indeces of the two vertices. The index must be in the range
            0 to number of vertices - 1.

        Returns
        -------
        is_edge : integer
            0 if vertex1 and vertex2 are not connected by a single edge,
            1 if vertex1 and vertex2 are connected by a boundary edge,
            2 if vertex1 and vertex2 are connected by an internal edge.
        """
        # plausibility test of vertex_index
        # is the index of type integer
        if ((not isinstance(vertex1_index, (int, np.integer))) or
                (not isinstance(vertex2_index, (int, np.integer)))):
            raise TypeError("""The parameters vertex1|2_index has to be
            int.""")

        # is the index within the interval (0, maxindex_vertlist)
        if (vertex1_index < 0) or \
           (vertex1_index > self.vertlist.shape[0]) or \
           (vertex2_index < 0) or \
           (vertex2_index > self.vertlist.shape[0]):
            raise IndexError("""The vertex1|2_index is out of range.""")

        # Determine the number of triangles in which the searched edge is
        # contained and return it.
        tri_with_edge = \
            self.trilist[np.logical_and((self.trilist ==
                                         vertex1_index).any(axis=1),
                                        (self.trilist ==
                                         vertex2_index).any(axis=1))]
        return len(tri_with_edge)

    def remove_vertices(self, vertex_index_list):
        """Remove a single vertex from a triangular mesh

        The method removes a single vertex from a triangle mesh.
        The Half-edge Collapse method is used.

        Parameters
        ----------
        vertex_index_list : vector of ints
            Indices of the vertices to remove from the mesh. The indices
            must be in the range 0 to number of vertices - 1.

        Returns
        -------
        triangsamples : trimesh object
            A trimesh object from the package spharapy, where the given
            vertex is removed.

        """
        # create a numpy array with unique indices to vertices to delete
        vertex_index_list = np.unique(vertex_index_list)
        # sort the index list in descending order
        vertex_index_list[::-1].sort()

        # plausibility test of vertex_index_list
        # is the index of type integer
        if vertex_index_list.ndim != 1:
            raise TypeError("""The parameter vertex_index_list has to be
            a 1D vector of int.""")

        if not all(isinstance(i, (int, np.integer))
                   for i in vertex_index_list):
            raise TypeError("""The parameter vertex_index_list has to be
            a 1D vector of int.""")

        # is the index within the interval (0, maxindex_vertlist)
        if ((vertex_index_list.min() < 0) or
                (vertex_index_list.max() > self.vertlist.shape[0])):
            raise IndexError("""The vertex_index is out of range.""")

        # create a deep copy
        temp_trimesh = copy.deepcopy(self)

        # iterate over the list of indices to be deleted
        for vertex_index in vertex_index_list:
            # determine all neighbor vertices of vertex in iteration
            neighbor_vert = temp_trimesh.one_ring_neighborhood(vertex_index)

            # determine all adjacent triangles of the given vertex
            adjacent_tri = temp_trimesh.adjacent_tri(vertex_index)

            # determine the type of edges to the neighbor vertices
            type_of_edges = \
                np.asarray([temp_trimesh.is_edge(vertex_index, vert_i)
                            for vert_i in neighbor_vert])

            # measure the 'distortion' of the mesh if a give edge is
            # 'half-colpsee', the measure is the deviation of the newly
            # generated triangles from equilateralism

            # 1st allocate mem
            dist_measure = np.zeros(len(neighbor_vert))

            # Iterate over all neighbour vertices of the vertex to be deleted
            for j, vert_i in enumerate(neighbor_vert):
                # Remove triangles containing the neighbour vertex
                temp_tri = \
                    adjacent_tri[(adjacent_tri != vert_i).all(axis=1), :]

                # Map the vertex to be deleted to the neighboring vertex
                temp_tri[temp_tri == vertex_index] = vert_i

                # Measure of distortion is the mean value of the variance
                # of the internal angles of the new triangles resulting
                # from the deletion of the vertex.
                dist_measure[j] = \
                    np.mean([np.var(angles_triangle(i[0], i[1], i[2]))
                             for i in temp_trimesh.vertlist[temp_tri]])

            # indexes of the neighboring vertices, sorted in ascending order
            # by distortion measure
            neighbor_vert_sort_dist = neighbor_vert[np.argsort(dist_measure)]

            # determine the index of the neighbor vertex with least
            # distortion that has exactly two neighbors in common with the
            # vertex to be deleted
            neighbor_vert_sel = \
                next(x for x in neighbor_vert_sort_dist
                     if len(np.intersect1d(neighbor_vert,
                            temp_trimesh.one_ring_neighborhood(x)))
                     == 2)

            # remove the two triangles, which contain both vertices
            temp_trimesh.trilist = \
                temp_trimesh.trilist[np.logical_not(
                    np.logical_and((temp_trimesh.trilist ==
                                    vertex_index).any(axis=1),
                                   (temp_trimesh.trilist ==
                                    neighbor_vert_sel).any(axis=1)))]

            # the selected neighbor vertex replaces the vertex to be deleted
            temp_trimesh.trilist = \
                np.where(temp_trimesh.trilist == vertex_index,
                         neighbor_vert_sel, temp_trimesh.trilist)

            # if the vertex to be deleted is located on the boundary of the
            # triangle mesh, insert a new triangle.
            if neighbor_vert[type_of_edges == 1].any():
                temp_trimesh.trilist = \
                    np.concatenate((temp_trimesh.trilist,
                                    [np.append(neighbor_vert[type_of_edges
                                                             == 1],
                                               neighbor_vert_sel)]))

            # remove vertex from vertex list
            temp_trimesh.vertlist = np.delete(temp_trimesh.vertlist,
                                              vertex_index, 0)

            # adjust indices of the triangle list
            temp_trimesh.trilist = \
                np.where(temp_trimesh.trilist < vertex_index,
                         temp_trimesh.trilist,
                         temp_trimesh.trilist - 1)

        # return the mesh without the vertices to be deleted
        return temp_trimesh

    def main():
        import trimesh as tm
        import datasets as sd

        mesh_in = sd.load_simple_triangular_mesh()
        vertlist = np.array(mesh_in['vertlist'])
        trilist = np.array(mesh_in['trilist'])

        testtrimesh = tm.TriMesh(trilist, vertlist)

        print("python main function")

    if __name__ == '__main__':
        main()



# -*- coding: utf-8 -*-
"""SPHARA filter

This module provides a class to perform a spatial filtering using a
SPHARA basis. The class is derived from
:class:`spharapy.spharabasis.SpharaBasis`. It provides methodes to
design different types of filters and to apply this filters to
spatially irregularly sampled data.

"""

import numpy as np
from scipy import linalg
from spharapy.spharabasis import SpharaBasis

__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.13"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


class SpharaFilter(SpharaBasis):
    """SPHARA filter class

    This class is used to design different types of filters and to
    apply this filters to spatially irregularly sampled data.

    Parameters
    ----------
    triangsamples : trimesh object
        A trimesh object from the package spharapy in which the triangulation
        of the spatial arrangement of the sampling points is stored. The SPHARA
        basic functions are determined for this triangulation of the sample
        points.
    mode : {'unit', 'inv_euclidean', 'fem'}, optional
        The discretisation method used to estimate the Laplace-Beltrami
        operator. Using the option 'unit' all edges of
        the mesh are weighted by unit weighting function. The option
        'inv_euclidean' results in edge weights corresponding to the
        inverse Euclidean distance of the edge lengths. The option
        'fem' uses a FEM discretisation. The default weighting
        function is 'fem'.
    specification : integer or array, shape (1, n_points)
        If an integer value for specification is passed to the
        constructor, it must be within the interval (-n_points,
        n_points), where n_points is the number of spatial sample
        points. If a positive integer value is passed, a spatial
        low-pass filter with the corresponding number of SPHARA basis
        functions is created, if a negative integer value is passed, a
        spatial low-pass filter is created. If a vector is passed,
        then all SPHARA basis functions corresponding to nonzero
        elements of the vector are used to create the filter. The
        default value of specification is 0, it means a neutral
        all-pass filter is designed and applied.

    """

    def __init__(self, triangsamples=None, mode='fem', specification=0):
        SpharaBasis.__init__(self, triangsamples, mode)
        self.specification = specification
        self._basis = None
        self._frequencies = None
        self._massmatrix = None
        self._filtermatrix = None

    @property
    def specification(self):
        """Get or set the specification of the filter.

        The parameter `specification` has to be an integer or a vector.
        Setting the `specification` will simultaneously apply a plausibility
        check.

        """

        return self._specification

    @specification.setter
    def specification(self, specification):
        if isinstance(specification, (int)):
            if np.abs(specification) > self._triangsamples.vertlist.shape[0]:
                raise ValueError("""The Number of selected basic functions is
                too large.""")
            else:
                if specification == 0:
                    self._specification = \
                        np.ones(self._triangsamples.vertlist.shape[0])
                else:
                    self._specification = \
                        np.zeros(self._triangsamples.vertlist.shape[0])
                    if specification > 0:
                        self._specification[:specification] = 1
                    else:
                        self._specification[specification:] = 1
        elif isinstance(specification, (list, tuple, np.ndarray)):
            specification = np.asarray(specification)
            if specification.shape[0] != self._triangsamples.vertlist.shape[1]:
                raise IndexError("""The length of the specification vector
                does not match the number of spatial sample points.""")
            else:
                self._specification = specification
        else:
            raise TypeError("""The parameter specification has to be
            int or a vecor.""")

    def filter(self, data):
        r"""Perform the SPHARA filtering

        This method performs the spatial SPHARA filtering
        for data defined at spatially distributed sampling points
        described by a triangular mesh. The filtering is
        performed by matrix multiplication of the data matrix and a
        precalculated filter matrix.

        Parameters
        ----------
        data : array, shape(m, n_points)
            A matrix with data to be filtered by spatial SPHARA
            filter. The number of vertices of the triangular mesh is
            n_points. The order of the spatial sample points must
            correspond to that in the vertex list used to determine
            the SPHARA basis functions.

        Returns
        -------
        data_filtered : array, shape (m, n_points)
            A matrix containing the filtered data.

        Examples
        --------

        >>> import numpy as np
        >>> from spharapy import trimesh as tm
        >>> from spharapy import spharafilter as sf
        >>> # define the simple test mesh
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> # create a spatial lowpass filter, FEM discretisation
        >>> sf_fem = sf.SpharaFilter(testtrimesh, mode='fem',
        ...                          specification=[1., 1., 0.])
        >>> # create some test data
        >>> data = np.concatenate([[[0., 0., 0.], [1., 1., 1.]],
        ...                          np.transpose(sf_fem.basis()[0])])
        >>> data
        array([[ 0.        ,  0.        ,  0.        ],
               [ 1.        ,  1.        ,  1.        ],
               [ 0.53452248,  0.53452248,  0.53452248],
               [-0.49487166, -0.98974332,  1.48461498],
               [ 1.42857143, -1.14285714, -0.28571429]])
        >>> # filter the test data
        >>> data_filtered = sf_fem.filter(data)
        >>> data_filtered
        array([[  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  1.00000000e+00,   1.00000000e+00,   1.00000000e+00],
               [  5.34522484e-01,   5.34522484e-01,   5.34522484e-01],
               [ -4.94871659e-01,  -9.89743319e-01,   1.48461498e+00],
               [ -1.69271249e-16,  -2.75762028e-16,   3.10220481e-16]])

        """

        # lazy evaluation, compute the basis at the first request and store
        # it until the triangular mesh or the discretization method is changed
        if self._basis is None or self._frequencies is None:
            self.basis()

        # lazy evaluation, compute the filter matrix at the first request and
        # store it until the triangular mesh or the discretization method is
        # changed
        if self._filtermatrix is None:
            basis_fun_sel = np.matmul(self._basis,
                                      np.diag(self._specification))
            if self._mode == 'fem':
                self._filtermatrix = np.matmul(
                    np.matmul(self._massmatrix, basis_fun_sel),
                    basis_fun_sel.transpose())
            else:
                self._filtermatrix = np.matmul(basis_fun_sel,
                                               basis_fun_sel.transpose())

        data = np.asarray(data)

        # Does the number of spatial samples of the data correspond to
        # that of the basis functions?
        if self._basis.shape[0] != data.shape[1]:
            raise ValueError('Dimension of data and filter matrix not equal.')

        # filter the data
        data_filtered = np.matmul(data, self._filtermatrix)

        return data_filtered



# -*- coding: utf-8 -*-
"""SPHARA basis functions

This module provides a class for determining SPHARA basis functions. Methodes
are provided to determine basis functions using different discretization
scheme of the Laplace-Beltrami operator, as FEM, inverse euclidean and unit.

"""

import numpy as np
from scipy import linalg
import spharapy.trimesh as tm

__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.13"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


class SpharaBasis(object):
    """SPHARA basis functions class

    This class can be used to determine SPHARA basis functions for spatially
    irregularly sampled functions whose topology is described by a triangular
    mesh.

    Parameters
    ----------
    triangsamples : trimesh object
        A trimesh object from the package spharapy in which the triangulation
        of the spatial arrangement of the sampling points is stored. The SPHARA
        basic functions are determined for this triangulation of the sample
        points.
    mode : {'unit', 'inv_euclidean', 'fem'}, optional
        The discretization method used to estimate the Laplace-Beltrami
        operator. Using the option 'unit' all edges of
        the mesh are weighted by unit weighting function. The option
        'inv_euclidean' results in edge weights corresponding to the
        inverse Euclidean distance of the edge lengths. The option
        'fem' uses a FEM discretization. the default weighting
        function is 'fem'.

    Attributes
    ----------
    triangsamples: trimesh object
        Triangulation of the spatial arrangement of the sampling points
    mode: {'unit', 'inv_euclidean', 'fem'}
        Discretization used to estimate the Laplace-Beltrami
        operator

    """

    def __init__(self, triangsamples=None, mode='fem'):
        self.triangsamples = triangsamples
        self.mode = mode
        self._basis = None
        self._frequencies = None
        self._massmatrix = None

    @property
    def triangsamples(self):
        """Get or set the triangsamples object.

        The parameter `triangsamples` has to be an instance of the
        class `spharapy.trimesh.TriMesh`. Setting the triangsamples
        object will simultaneously check if it in the correct format.

        """

        return self._triangsamples

    @triangsamples.setter
    def triangsamples(self, triangsamples):
        if not isinstance(triangsamples, tm.TriMesh):
            raise TypeError('triangsamples is no instance of TriMesh')
        # pylint: disable=W0201
        self._triangsamples = triangsamples
        self._basis = None
        self._frequencies = None
        self._massmatrix = None

    @property
    def mode(self):
        """Get or set the discretization method.

        The discretization method used to estimate the Laplace-Beltrami
        operator, choosen from {'unit', 'inv_euclidean', 'fem'}. Setting
        the triangsamples object will simultaneously check if it in the
        correct format.

        """

        return self._mode

    @mode.setter
    def mode(self, mode):
        # plausibility test of option 'mode'
        if mode not in ('unit', 'inv_euclidean', 'fem'):
            raise ValueError("Unrecognized mode '%s'" % mode)
        # pylint: disable=W0201
        self._mode = mode
        self._basis = None
        self._frequencies = None
        self._massmatrix = None

    def basis(self):
        """Return the SPHARA basis for the triangulated sample points

        The method determines a SPHARA basis for spatially distributed
        sampling points described by a triangular mesh. A discrete
        Laplace-Beltrami operator in matrix form is determined for the
        given triangular grid. The discretization methods for
        determining the Laplace-Beltrami operator is specified in the
        attribute `mode`. The eigenvectors :math:`\\vec{x}` and the
        eigenvalues :math:`\\lambda` of the matrix :math:`L`
        containing the discrete Laplace-Beltrami operator are the
        SPHARA basis vectors and the natural frequencies,
        respectively, :math:`L \\vec{x} = \\lambda \\vec{x}`.

        Parameters
        ----------

        Returns
        -------
        basis : array, shape (n_points, n_points)
            Matrix, which contains the SPHARA basis functions column by column.
            The number of vertices of the triangular mesh is n_points.
        frequencies : array, shape (n_points, 1)
            The natural frequencies associated to the SPHARA basis functions.

        Examples
        --------

        >>> from spharapy import trimesh as tm
        >>> from spharapy import spharabasis as sb
        >>> testtrimesh = tm.TriMesh([[0, 1, 2]], [[1., 0., 0.], [0., 2., 0.],
        ...                                        [0., 0., 3.]])
        >>> sb_fem = sb.SpharaBasis(testtrimesh, mode='fem')
        >>> sb_fem.basis()
        (array([[ 0.53452248, -0.49487166,  1.42857143],
                [ 0.53452248, -0.98974332, -1.14285714],
                [ 0.53452248,  1.48461498, -0.28571429]]),
         array([  2.33627569e-16,   1.71428571e+00,   5.14285714e+00]))

        """

        # lazy evaluation, compute the basis at the first request and store
        # it until the triangular mesh or the discretization method is changed
        if self._basis is None or self._frequencies is None:
            if self.mode == 'fem':
                self._massmatrix = (self.triangsamples
                                    .massmatrix(mode='normal'))
                stiffmatrix = self.triangsamples.stiffnessmatrix()
                self._frequencies, self._basis = linalg.eigh(-stiffmatrix,
                                                             self._massmatrix)
                # self._basis =
            else:  # 'unit' and 'inv_euclidean' discretization
                laplacianmatrix = (self.triangsamples
                                   .laplacianmatrix(mode=self.mode))
                self._frequencies, self._basis = linalg.eigh(laplacianmatrix)

        # make a row vector of natural frequencies
        # print(self._frequencies)
        # self._frequencies = self._frequencies.transpose
        # print(self._frequencies.shape)
        # return the SPHARA basis
        return self._basis, self._frequencies

    def massmatrix(self):
        """Return the massmatrix

        The method returns the mass matrix of the triangular mesh.

        """
        # lazy evaluation, compute the mass matrix at the first request and
        # store it until the triangular mesh or the discretization method
        # is changed
        if self._massmatrix is None:
            self._massmatrix = self.triangsamples.massmatrix(mode='normal')

        return self._massmatrix




# coding: utf-8

# In[1]:


__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.0"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')
from IPython.display import set_matplotlib_formats
# set_matplotlib_formats('pdf', 'svg')
set_matplotlib_formats('png')


# In[3]:


get_ipython().system('jt -r')


# # Introduction to spatial harmonic analysis of EEG data using SpharaPy

# Training course 4: EEG/MEG - Spatial filters and uncertainty analysis
# 
# 4th International BaCI Conference 2019
# 
# Uwe Graichen, TU Ilmenau, BMTI
# 
# Chengdu, 2019/09/11

# ## Spatial Fourier analysis - Motivation
# 
# <img src="images/2D_Fourier_Example.svg.png" alt="2DFourier" style="width: 100%;"/>
# 
# - Fourier analysis is a standard tool in digital signal and image processing
#     - Filtering, Data compression
#     - Compressed Sensing, Acceleration of processing, ...
# - Common digital images: flat domain, sampling points on Cartesian or rectangular grid
# - For 2D Fourier transform on a Cartesian grid, Fourier basis is given inherently
# $F[k,l] = \sum_{m=0}^{M-1} \sum_{n=0}^{N-1} f[m,n] \cdot
#       \color{red}{\mathrm{e}^{-2\pi \mathrm{i} \cdot \left(\frac{mk}{M}  +
#             \frac{nl}{N} \right) }}$

# ## Spatial Fourier analysis - EEG Data
# 
# <img src="images/EEG_image-1.png" alt="EEG_image" style="width: 90%;"/>
# 
# - Common digital images:
#     - flat domain
#     - sampling points on Cartesian or rectangular grid
# - Diskret sampled potential distribution:
#     - curved domain
#     - irregular sampling data (sample points are not on Cartesian grid)
#     - Fourier basis cannot be given inherently $F[k,l] = \sum_{m=0}^{M-1} \sum_{n=0}^{N-1} f[m,n] \cdot
#       \color{red}{{\text{<suitable Fourier basis>}}}$ 
# 

# ## Spatial Fourier analysis - 2D image data
# - $$F[k,l] = \sum_{m=0}^{M-1} \sum_{n=0}^{N-1} f[m,n] \cdot
#       \color{red}{\mathrm{e}^{-2\pi \mathrm{i} \cdot \left(\frac{mk}{M}  +
#             \frac{nl}{N} \right) }}$$
# - Fourier basis (a set of harmonic functions) is solution of Laplacian eigenvalue problem (Helmholtz equation) $$L \vec{x} = \lambda \vec{x}$$
# - $L \in \mathbb{R}^{M \times N}$ discrete Laplace(-Beltrami) operator of the image domain, matrix notation
# - $\vec{x}$ eigenvector, harmonic function (basis function)
# - $\lambda$ eigenvalue, natural frequency

# ## Generalization of spatial Fourier analysis
# 
# <img src="images/Sensor_tria_mesh-1.png" alt="EEG_image" style="width: 90%;"/>
# 
# - Triangulation of sensor positions
# - Definition of a discrete Laplace(-Beltrami) operator for triangular mesh, Laplace matrix $L$
# - Solving the Laplacian eigenvalue problem $$L \vec{x} = \lambda \vec{x}$$
# - Eigenvectors form a harmonic, orthonormal basis, applicable for spatial Fourier analysis
# - **SPHARA** – **SP**atial **HAR**monic **A**nalysis

# ## Generalization of spatial Fourier analysis
# 
# - Solving the Laplacian eigenvalue problem $$L \vec{x} = \lambda \vec{x}$$
# - Eigenvectors form a harmonic, orthonormal basis, applicable for spatial Fourier analysis
# 
# <img src="images/Low_frequ_bf-1.png" alt="Low_freq_bf" style="width: 90%;"/>
# 
# - **SPHARA** – **SP**atial **HAR**monic **A**nalysis

# ## SpharaPy
# 
# - Python implementation of a new approach for spatial harmonic analysis (SPHARA)
# - SpharaPy Python toolbox provides classes and functions:
#     - to compute the SPHARA BF
#     - for data analysis and synthesis
#     - to design and apply spatial filters
# <img src="images/Class_diag.png" alt="Class_diagram" style="width: 80%;"/>

# ## Import the SpharaPy Python packages
# 
# - At the beginning we import three modules of the SpharaPy package as well as several other packages and single functions from packages.

# In[4]:


# import modules from spharapy package
import spharapy.trimesh as tm
import spharapy.spharatransform as st
import spharapy.spharafilter as sf
import spharapy.datasets as sd

# import additional modules used in this tutorial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.offsetbox import AnchoredText
from matplotlib.text import Text
from matplotlib.colors import LogNorm
from matplotlib.patches import Circle, Rectangle
import numpy as np


# ## Load example data set
# 
# - SEP data of a single subject recorded with a 256 channel EEG system with equidistant layout
# - The data set is one of the example data sets contained in the SpharaPy toolbox

# In[5]:


# loading the 256 channel EEG dataset from spharapy sample datasets
data_in = sd.load_eeg_256_channel_study()


# - The dataset includes lists of vertices, triangles, and sensor labels, as well as EEG data (somatosensory-evoked potentials SEP)

# In[6]:


print(data_in.keys())


# ## Load example data set

# - Triangulation, 256 vertices and 482 triangles
# - 256 channels and 369 time samples, 50 ms before to 130 ms after stimulation
# - Sampling frequency is 2048 Hz.

# In[7]:


vertlist = np.array(data_in['vertlist'])
trilist = np.array(data_in['trilist'])
eegdata = np.array(data_in['eegdata'])
print('vertices = ', vertlist.shape)
print('triangles = ', trilist.shape)
print('eegdata = ', eegdata.shape)


# ## Example data set
# 
# - Visualization of triangulated sensor setup

# In[8]:


fig = plt.figure(dpi = 150)
fig.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)
ax = fig.gca(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')
ax.set_title('The triangulated EEG sensor setup')
ax.view_init(elev=20., azim=80.)
ax.set_aspect('auto')

ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1], vertlist[:, 2],
                triangles=trilist, color='lightblue', edgecolor='black',
                linewidth=0.5, shade=True)
plt.show()


# ## Example data set
# 
# - Visualization of the averaged SEP data as a butterfly plot. The components P14, N20 and N30 are marked by red, blue and green vertical lines

# In[9]:


x = np.arange(-50, 130, 1/2.048)
figeeg = plt.figure(dpi = 150)
axeeg = figeeg.gca()
axeeg.axvline(13, color='red')
axeeg.axvline(19, color='blue')
axeeg.axvline(30, color='green')
axeeg.plot(x, eegdata[:, :].transpose())
axeeg.set_xlabel('t/ms')
axeeg.set_ylabel('V/µV')
axeeg.set_title('SEP data')
axeeg.set_ylim(-3.5, 3.5)
axeeg.set_xlim(-50, 130)
axeeg.grid(True)
plt.show()


# ## SPHARA analysis of EEG data
# ### Create a SpharaPy TriMesh instance

# In[10]:


# create an instance of the TriMesh class
mesh_eeg = tm.TriMesh(trilist, vertlist)


# ### Create a SpharaPy SpharaTransform instance

# In[11]:


sphara_transform_fem = st.SpharaTransform(mesh_eeg, 'fem')
basis_functions_fem, natural_frequencies_fem     = sphara_transform_fem.basis()


# ### Visualization the basis functions
# 
# - The first 12 spatially low-frequency SPHARA BF of the basis used for the transform

# In[12]:


figsb1, axes1 = plt.subplots(nrows=3, ncols=4, figsize=(6.0, 4.4),dpi = 120,
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_fem[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    if i == 0:
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
    else:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_zticklabels([])
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('auto')
    ax.add_artist(AnchoredText('BF' + str(i), frameon=True, loc='upper center'))
    # ax.text2D(0, 0.8, 'local max')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.autoscale()
    trisurfplot.set_clim(-0.01, 0.01)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.95,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.45, 0))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.15, top=1.0, wspace=0.0, hspace=0.2)
plt.show()


# ### SPHARA transform of the EEG data
# 
# <img src="images/scheme_sphara_decomposition.svg" alt="SPHARA_Decomp" style="width: 100%;"/>

# ### SPHARA transform of the EEG data
# 
# <img src="images/scheme_sphara_decomposition2.svg" alt="SPHARA_Decomp" style="width: 100%;"/>

# ### SPHARA transform of the EEG data
# 
# - Perform the SPHARA transformation of the EEG data

# In[13]:


# perform the SPHARA transform
sphara_trans_eegdata =     sphara_transform_fem.analysis(eegdata.transpose())


# ### SPHARA transform of the EEG data
# 
# - Butterfly plot of all channels of the EEG
# - Power contributions of the first 60 SPHARA basis functions

# In[14]:


# 60 low-frequency basis functions are displayed
ysel = 60
figsteeg, (axsteeg1, axsteeg2) = plt.subplots(nrows=2, figsize=(6.0, 5.0), dpi = 100)

y = np.arange(0, ysel)
x = np.arange(-50, 130, 1/2.048)

axsteeg1.axvline(13, color='red')
axsteeg1.axvline(19, color='blue')
axsteeg1.axvline(30, color='green')
axsteeg1.plot(x, eegdata[:, :].transpose())
axsteeg1.set_ylabel('V/µV')
axsteeg1.set_title('EEG data, 256 channels')
axsteeg1.set_ylim(-2.5, 2.5)
axsteeg1.set_xlim(-50, 130)
axsteeg1.grid(True)
axsteeg1.add_artist(AnchoredText("a", frameon=True, loc='upper left'))

pcm2 = axsteeg2.pcolormesh(x, y,
                          np.square(np.abs(sphara_trans_eegdata.transpose()
                                           [0:ysel, :])),
                         norm=LogNorm(vmin=0.1,vmax=50000),
                         linewidth=0.0,rasterized=True)
axsteeg2.set_xlabel('t/ms')
axsteeg2.set_ylabel('# BF')
axsteeg2.set_title('Power contribution of SPHARA basis functions')
axsteeg2.grid(True)
axsteeg2.add_artist(AnchoredText("b", frameon=True, loc='upper left'))
axsteeg2.add_artist(AnchoredText("logarithmic color map", frameon=True, loc='upper right'))
figsteeg.colorbar(pcm2, ax=[axsteeg2], shrink=0.9,
                  anchor=(0.85, 0.0), label='power / a.u.')

plt.subplots_adjust(left=0.1, right=0.85, bottom=0.1, top=0.95, hspace=0.3)
plt.show()


# ## SPHARA filter using FEM discretisation

# ### Create a SpharaPy SpharaFilter instance
# 
# - Determine an instance of the class SpharaFilter, which is used to execute the spatial filtering
# - use a Laplace-Beltrami operator, which is discretized by the FEM approach

# In[15]:


sphara_filter_fem = sf.SpharaFilter(mesh_eeg, mode='fem',
                                    specification=20)


# ### SPHARA filtering of the EEG data
# 
# <img src="images/scheme_sphara_filtering.svg" alt="SPHARA_Filter" style="width: 100%;"/>

# ### SPHARA filtering of the EEG data
# 
# - Perform the SPHARA filtering of the EEG data

# In[16]:


# perform the SPHARA filtering
sphara_filt_eegdata =     sphara_filter_fem.filter(eegdata.transpose()).transpose()


# ### SPHARA filtering of the EEG data

# In[17]:


figsteeg, (axsteeg1, axsteeg2) = plt.subplots(nrows=2, figsize=(6.0, 6.0), dpi = 100)

axsteeg1.axvline(13, color='red')
axsteeg1.axvline(19, color='blue')
axsteeg1.axvline(30, color='green')
axsteeg1.plot(x, eegdata[:, :].transpose())
axsteeg1.set_title('Unfiltered EEG data')
axsteeg1.set_ylabel('V/µV')
axsteeg1.set_xlabel('t/ms')
axsteeg1.set_ylim(-2.5, 2.5)
axsteeg1.set_xlim(-50, 130)
axsteeg1.grid(True)

axsteeg2.axvline(13, color='red')
axsteeg2.axvline(19, color='blue')
axsteeg2.axvline(30, color='green')
axsteeg2.plot(x, sphara_filt_eegdata[:, :].transpose())
axsteeg2.set_title('SPHARA low-pass filtered EEG data, 20 BF, fem')
axsteeg2.set_ylabel('V/µV')
axsteeg2.set_xlabel('t/ms')
axsteeg2.set_ylim(-2.5, 2.5)
axsteeg2.set_xlim(-50, 130)
axsteeg2.grid(True)

plt.subplots_adjust(left=0.1, right=0.95, bottom=0.1, top=0.95, hspace=0.3)
plt.show()


# ### SPHARA filtering of the EEG data

# In[18]:


time_pts = [129, 141, 164]
figsf1, axessf1 = plt.subplots(nrows=2, ncols=3, figsize=(6.0, 4.4), dpi = 100,
                               subplot_kw={'projection': '3d'})

for i in range(2):
    for j in range(3):
        if i == 0:
            colorssf1 = np.mean(eegdata[trilist, time_pts[j]], axis=1)
        else:
            colorssf1 = np.mean(sphara_filt_eegdata[trilist, time_pts[j]],
                                axis=1)
        
        axessf1[i, j].set_xlabel('x')
        axessf1[i, j].set_ylabel('y')
        axessf1[i, j].set_zlabel('z')
        axessf1[i, j].view_init(elev=60., azim=80.)
        axessf1[i, j].set_aspect('auto')

        trisurfplot = axessf1[i, j].plot_trisurf(vertlist[:, 0],
                                                 vertlist[:, 1],
                                                 vertlist[:, 2],
                                                 triangles=trilist,
                                                 cmap=plt.cm.bwr,
                                                 edgecolor='white',
                                                 linewidth=0.)
        trisurfplot.set_array(colorssf1)
        trisurfplot.set_clim(-2., 2)

cbar = figsb1.colorbar(trisurfplot, ax=axessf1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, 0.0))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.225, top=0.95,
                    wspace=0.0, hspace=0.0)
plt.show()


# ### Application of SPHARA filter to data with artificial noise
# 
# - EEG data are disturbed by white noise with SNR 3dB, 0dB and -3dB
# - Spatial low-pass SPHARA filter with 20 basis functions

# In[19]:


# vector with noise levels in dB
db_val_vec = [3, 0, -3]

# compute the power of the SEP data
power_sep = np.sum(np.square(np.absolute(eegdata))) / eegdata.size

# compute a vector with standard deviations of the noise
# in relation to signal power for the given noise levels
noise_sd_vec = list(map(lambda db_val: 
                        np.sqrt(power_sep / (10 ** (db_val / 10))),
                        db_val_vec))

# add the noise to the EEG data
eegdata_noise = list(map(lambda noise_sd: eegdata + 
                         np.random.normal(0, noise_sd, [256, 369]),
                         noise_sd_vec))

# filter the EEG data containing the artificial noise
eegdata_noise_filt = list(map(lambda eeg_noise:
                    (sphara_filter_fem.filter(eeg_noise.transpose()).
                     transpose()), eegdata_noise))


# ### Application of SPHARA filter to data with artificial noise

# In[20]:


figfilt, axesfilt = plt.subplots(nrows=4, ncols=2, figsize=(6.0, 6.8), dpi = 100)

axesfilt[0, 0].axvline(13, color='red')
axesfilt[0, 0].axvline(19, color='blue')
axesfilt[0, 0].axvline(30, color='green')
axesfilt[0, 0].plot(x, eegdata[:, :].transpose())
axesfilt[0, 0].set_title('EEG')
axesfilt[0, 0].set_ylabel('V/µV')
# axesfilt[0, 0].set_xlabel('t/ms')
axesfilt[0, 0].set_ylim(-2.5, 2.5)
axesfilt[0, 0].set_xlim(-50, 130)
axesfilt[0, 0].grid(True)

axesfilt[0, 1].axvline(13, color='red')
axesfilt[0, 1].axvline(19, color='blue')
axesfilt[0, 1].axvline(30, color='green')
axesfilt[0, 1].plot(x, sphara_filt_eegdata[:, :].transpose())
axesfilt[0, 1].set_title('SPHARA low-pass filtered EEG')
# axesfilt[0, 1].set_ylabel('V/µV')
# axesfilt[0, 1].set_xlabel('t/ms')
axesfilt[0, 1].set_ylim(-2.5, 2.5)
axesfilt[0, 1].set_xlim(-50, 130)
axesfilt[0, 1].grid(True)

for i in range(3):
    axesfilt[i + 1, 0].axvline(13, color='red')
    axesfilt[i + 1, 0].axvline(19, color='blue')
    axesfilt[i + 1, 0].axvline(30, color='green')
    axesfilt[i + 1, 0].plot(x, eegdata_noise[i].transpose())
    
    axesfilt[i + 1, 0].set_title('EEG + noise, SNR ' +
                                 str(db_val_vec[i]) + 'dB')
    axesfilt[i + 1, 0].set_ylabel('V/µV')
    if i == 2:
        axesfilt[i + 1, 0].set_xlabel('t/ms')
    axesfilt[i + 1, 0].set_ylim(-2.5, 2.5)
    axesfilt[i + 1, 0].set_xlim(-50, 130)
    axesfilt[i + 1, 0].grid(True)

    axesfilt[i + 1, 1].axvline(13, color='red')
    axesfilt[i + 1, 1].axvline(19, color='blue')
    axesfilt[i + 1, 1].axvline(30, color='green')
    axesfilt[i + 1, 1].plot(x, eegdata_noise_filt[i].transpose())
    axesfilt[i + 1, 1].set_title('EEG + noise, SNR ' +
                                 str(db_val_vec[i]) + 'dB, SPHARA filtered')
    # axesfilt[i + 1, 1].set_ylabel('V/µV')
    if i == 2:
        axesfilt[i + 1, 1].set_xlabel('t/ms')
    axesfilt[i + 1, 1].set_ylim(-2.5, 2.5)
    axesfilt[i + 1, 1].set_xlim(-50, 130)
    axesfilt[i + 1, 1].grid(True)

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.05, top=0.95, hspace=0.35,
                   wspace=0.15)
plt.show()


# ### Application of SPHARA filter to data with artificial noise

# In[21]:


time_pts = [129, 141, 164]
figfilttopo, axesfilttopo = plt.subplots(nrows=4, ncols=6, figsize=(6.0, 4.4),
                                         dpi = 100, subplot_kw={'projection': '3d'})

for i in range(2):
    for j in range(3):

        if i == 0:
            colorssf1 = np.mean(eegdata[trilist, time_pts[j]], axis=1)
        else:
            colorssf1 = np.mean(sphara_filt_eegdata[trilist, time_pts[j]],
                                axis=1)

        #axesfilttopo[0, 3 * i + j].set_xlabel('x')
        #axesfilttopo[0, 3 * i + j].set_ylabel('y')
        #axesfilttopo[0, 3 * i + j].set_zlabel('z')
        axesfilttopo[0, 3 * i + j].view_init(elev=60., azim=80.)
        axesfilttopo[0, 3 * i + j].set_aspect('auto')
        axesfilttopo[0, 3 * i + j].grid(False)
        axesfilttopo[0, 3 * i + j].set_axis_off()
        axesfilttopo[0, 3 * i + j].dist = 6
        axesfilttopo[0, 3 * i + j].text(120,0,10,'F')
        axesfilttopo[0, 3 * i + j].text(-100,0,10,'O')
        axesfilttopo[0, 3 * i + j].text(0,80,10,'L')
        #axesfilttopo[0, 3 * i + j].text(0,-80,10,'R')


        trisurfplot = axesfilttopo[0, 3 * i + j].plot_trisurf(vertlist[:, 0],
                                                              vertlist[:, 1],
                                                              vertlist[:, 2],
                                                              triangles=trilist,
                                                              cmap=plt.cm.bwr,
                                                              edgecolor='white',
                                                              linewidth=0.)
        trisurfplot.set_array(colorssf1)
        trisurfplot.set_clim(-2., 2)
        
for i in range(3):
    for j in range(3):
        color_eeg_noise = np.mean(eegdata_noise[i][trilist, time_pts[j]], axis=1)
        color_eeg_filt = np.mean(eegdata_noise_filt[i][trilist, time_pts[j]], axis=1)
        
        #axesfilttopo[(i + 1), j].set_xlabel('x')
        #axesfilttopo[(i + 1), j].set_ylabel('y')
        #axesfilttopo[(i + 1), j].set_zlabel('z')
        axesfilttopo[(i + 1), j].view_init(elev=60., azim=80.)
        axesfilttopo[(i + 1), j].set_aspect('auto')
        axesfilttopo[(i + 1), j].grid(False)
        axesfilttopo[(i + 1), j].set_axis_off()
        axesfilttopo[(i + 1), j].dist = 6
        axesfilttopo[(i + 1), j].text(120,0,10,'F')
        axesfilttopo[(i + 1), j].text(-100,0,10,'O')
        axesfilttopo[(i + 1), j].text(0,80,10,'L')
        #axesfilttopo[(i + 1), j].text(0,-80,10,'R')

        
        trisurfplot = axesfilttopo[(i + 1), j].plot_trisurf(vertlist[:, 0],
                                                              vertlist[:, 1],
                                                              vertlist[:, 2],
                                                              triangles=trilist,
                                                              cmap=plt.cm.bwr,
                                                              edgecolor='white',
                                                              linewidth=0.)
        trisurfplot.set_array(color_eeg_noise)
        trisurfplot.set_clim(-2., 2)

        #axesfilttopo[(i + 1), 3 + j].set_xlabel('x')
        #axesfilttopo[(i + 1), 3 + j].set_ylabel('y')
        #axesfilttopo[(i + 1), 3 + j].set_zlabel('z')
        axesfilttopo[(i + 1), 3 + j].view_init(elev=60., azim=80.)
        axesfilttopo[(i + 1), 3 + j].dist=10
        axesfilttopo[(i + 1), 3 + j].set_aspect('auto')
        axesfilttopo[(i + 1), 3 + j].grid(False)
        axesfilttopo[(i + 1), 3 + j].set_axis_off()
        axesfilttopo[(i + 1), 3 + j].dist = 6
        axesfilttopo[(i + 1), 3 + j].text(120,0,10,'F')
        axesfilttopo[(i + 1), 3 + j].text(-100,0,10,'O')
        axesfilttopo[(i + 1), 3 + j].text(0,80,10,'L')
        #axesfilttopo[(i + 1), 3 + j].text(0,-80,10,'R')

        trisurfplot = axesfilttopo[(i + 1), 3 + j].plot_trisurf(vertlist[:, 0],
                                                              vertlist[:, 1],
                                                              vertlist[:, 2],
                                                              triangles=trilist,
                                                              cmap=plt.cm.bwr,
                                                              edgecolor='white',
                                                              linewidth=0.)
        trisurfplot.set_array(color_eeg_filt)
        trisurfplot.set_clim(-2., 2)

cbar = figfilttopo.colorbar(trisurfplot, ax=axesfilttopo.ravel().tolist(), shrink=0.95,
                       orientation='horizontal', label='V/µV', fraction=0.05, pad=0.05,
                       anchor=(0.49, 0.0))

figfilttopo.add_artist(Rectangle((0.010, 0.18), width=0.155, height=0.87, fill=False, color='red', lw=2))
figfilttopo.add_artist(Rectangle((0.178, 0.18), width=0.155, height=0.87, fill=False, color='blue', lw=2))
figfilttopo.add_artist(Rectangle((0.346, 0.18), width=0.155, height=0.87, fill=False, color='green', lw=2))
figfilttopo.add_artist(Rectangle((0.515, 0.18), width=0.155, height=0.87, fill=False, color='red', lw=2))
figfilttopo.add_artist(Rectangle((0.683, 0.18), width=0.155, height=0.87, fill=False, color='blue', lw=2))
figfilttopo.add_artist(Rectangle((0.851, 0.18), width=0.155, height=0.87, fill=False, color='green', lw=2))
figfilttopo.add_artist(Text(x=0.26, y=1.0, text='EEG', ha='center', 
                            bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.76, y=1.0, text='SPHARA low-pass filtered EEG',
                            ha='center', bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.26, y=0.79, text='EEG + noise, SNR 3 dB', ha='center',
                            bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.76, y=0.79, text='EEG + noise, SNR 3 dB, SPHARA filtered',
                            ha='center', bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.26, y=0.58, text='EEG + noise, SNR 0 dB', ha='center',
                            bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.76, y=0.58, text='EEG + noise, SNR 0 dB, SPHARA filtered',
                            ha='center', bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.26, y=0.365, text='EEG + noise, SNR -3 dB', ha='center',
                            bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))
figfilttopo.add_artist(Text(x=0.76, y=0.365, text='EEG + noise, SNR -3 dB, SPHARA filtered',
                            ha='center', bbox=dict(facecolor='white', edgecolor='white', pad=1.0)))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.2, top=1.0, hspace=0.3,
                   wspace=0.05)

plt.show()


# ## SpharaPy &ndash; Further useful sources of information
# ### Publication
# - Graichen U., Eichardt R., Fiedler P., Strohmeier D., Zanow F., Haueisen J.
# *SPHARA - A generalized spatial Fourier analysis for multi-sensor systems with non-uniformly arranged sensors: Application to EEG*, PLoS One, 10 (2015), pp. 1-22, https://doi.org/10.1371/journal.pone.0121741
# - Graichen U., Eichardt R., Haueisen J. *SpharaPy: A Python toolbox for spatial harmonic analysis of non-uniformly sampled data*, SoftwareX, Vol. 10, 2019, https://doi.org/10.1016/j.softx.2019.100289

# ## SpharaPy &ndash; Further useful sources of information
# ### Software etc.
# - Developer documentation/manual/tutorials: https://spharapy.readthedocs.io/
# - Code repository: https://gitlab.com/uwegra/spharapy
# - SpharaPy on Python Package Index (PyPI): https://pypi.org/project/SpharaPy/
# - This presentation: https://cloud.tu-ilmenau.de/s/Kc3bf9PzA6jBJeE



# -*- coding: utf-8 -*-
"""
Base IO code to provide sample datasets
"""

import os
import csv

import numpy as np

__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.13"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


def load_minimal_triangular_mesh():
    """Returns the triangulation of a single triangle

    The data set consists of a list of three vertices at the unit vectors
    of vector space :math:`\mathbb{R}^3`and a list of a single
    triangle.

    ===================    =
    Number of vertices     3
    Number of triangles    1
    ===================    =

    Parameters
    ----------
    None

    Returns
    -------
    triangulation : dictionary
        Dictionary-like object containing the triangulation of a single
        triangle. The attributes are: 'vertlist', the list of vertices,
        'trilist', the list of triangles.
    """
    return {'vertlist': [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            'trilist': [[0, 1, 2]]}


def load_simple_triangular_mesh():
    """Returns the triangulation of a simple triangular mesh

    The data set consists of a triangulation of an unit hemisphere.

    ===================    ===
    Number of vertices     131
    Number of triangles    232
    ===================    ===

    Parameters
    ----------
    None

    Returns
    -------
    triangulation : dictionary
        Dictionary-like object containing the triangulation of a
        simple triangular mesh. The attributes are: 'vertlist', the
        list of vertices, 'trilist', the list of triangles.
    """
    # import vertices data
    path_to_vertdata = os.path.join(os.path.dirname(__file__),
                                    'data/simple_mesh_vert.csv')
    with open(path_to_vertdata, 'r') as f:
        reader = csv.reader(f)
        datavertices = [[float(y) for y in x]
                        for x in list(reader)]
        f.close()

    # import of triangle list
    path_to_tridata = os.path.join(os.path.dirname(__file__),
                                   'data/simple_mesh_tri.csv')
    with open(path_to_tridata, 'r') as f:
        reader = csv.reader(f)
        datatriangles = [[int(y) for y in x]
                         for x in list(reader)]
        f.close()

    return {'vertlist': datavertices,
            'trilist': datatriangles}


def load_eeg_256_channel_study():
    """Load sensor setup and measured EEG data

    The data set consists of a triangulation of a 256 channel
    equidistant EEG cap and EEG data from previously performed
    experiment addressing the cortical activation related to
    somatosensory-evoked potentials (SEP). During the experiment the
    median nerve of the right forearm was stimulated by bipolar
    electrodes (stimulation rate: 3.7 Hz, interstimulus interval: 270
    ms, stimulation strength: motor plus sensor threshold
    :cite:`mauguiere99,cruccu08`, constant current rectangular pulse
    wave impulses with a length of 50 \mu s, number of stimulations:
    6000). Data were sampled at 2048 Hz and software high-pass (24
    dB/oct, cutoff-frequency 2 Hz) and notch (50 Hz and two harmonics)
    filtered. All trials were manually checked for artifacts, the
    remaining trials were averaged, see also S1 data set in
    :cite:`graichen15`.

    ===================    ========================================
    Number of vertices     256
    Number of triangles    480
    SEP Data (EEG)         256 channels, 369 time samples
    Time range             50 ms before to 130 ms after stimulation
    Sampling frequency     2048 Hz
    ===================    ========================================

    Parameters
    ----------
    None

    Returns
    -------
    triangulation and EEG data: dictionary
        Dictionary-like object containing the triangulation of a
        simple triangular mesh. The attributes are: 'vertlist', the
        list of vertices, 'trilist', the list of triangles,
        'labellist' the list of labels of the EEG channels, 'eegdata',
        an array containing the EEG data.
    """
    # import vertices data
    path_to_vertdata = os.path.join(os.path.dirname(__file__),
                                    'data/eeg_256_channels_vert.csv')
    with open(path_to_vertdata, 'r') as f:
        reader = csv.reader(f)
        datavertices = [[float(y) for y in x]
                        for x in list(reader)]
        f.close()

    # import of triangle list
    path_to_tridata = os.path.join(os.path.dirname(__file__),
                                   'data/eeg_256_channels_tri.csv')
    with open(path_to_tridata, 'r') as f:
        reader = csv.reader(f)
        datatriangles = [[int(y) for y in x]
                         for x in list(reader)]
        f.close()

    # import the sensor labels
    path_to_labeldata = os.path.join(os.path.dirname(__file__),
                                     'data/eeg_256_channels_label.csv')
    with open(path_to_labeldata, 'r') as f:
        reader = csv.reader(f)
        datalabels = list(reader)
        f.close()

    # import SEP EEG data
    path_to_eegdata = os.path.join(os.path.dirname(__file__),
                                   'data/eeg_256_channels_sep_data.csv')
    with open(path_to_eegdata, 'r') as f:
        reader = csv.reader(f)
        eegdata = [[float(y) for y in x]
                   for x in list(reader)]
        eegdata = np.asarray(eegdata)
        f.close()

    return {'vertlist': datavertices,
            'trilist': datatriangles,
            'labellist': datalabels,
            'eegdata': eegdata}


if __name__ == "__main__":
    minimesh = load_minimal_triangular_mesh()
    print((minimesh['trilist']))
    print((minimesh['vertlist']))

    simplemesh = load_simple_triangular_mesh()
    print((simplemesh['trilist']))
    print((simplemesh['vertlist']))

    sepstudy = load_eeg_256_channel_study()
    print((sepstudy['trilist']))
    print((sepstudy['vertlist']))
    print((sepstudy['labellist']))
    print((sepstudy['eegdata'].shape))



# -*- coding: utf-8 -*-

r""".. _spharapy_example:


"""

# # Short tutorial how to use the Python package SpharaPy

# ## Abstract

# This short tutorial presents some of the essential functions of the SpharaPy package, such as determining a SPHARA basis for a given spatial sensor setup, spatial-harmonic analysis using the SPHARA basis, and SPHARA-based spatial filtering.
# 
# The code examples from the article *SpharaPy: A Python toolbox for spatial harmonic analysis of non-uniformly sampled data* in the journal __[SoftwareX](https://www.sciencedirect.com/journal/softwarex)__ are taken from this tutorial. All figures from the above article can be reproduced with the help of this tutorial.
# 
# The tutorial is available as IPython in a Jupyter notebook and as Python source code.

# ## Import modules and functions

# At the beginning we import three modules of the SpharaPy package as
# well as several other packages and single functions.

# In[1]:


# uncomment the following line to allow interactive figures
get_ipython().magic('matplotlib notebook')

# uncomment the following lines to export the tutorial
# in PDF-Format with vector graphics
# %matplotlib inline
# from IPython.display import set_matplotlib_formats
# set_matplotlib_formats('png', 'pdf')

# import modules from spharapy package
import spharapy.trimesh as tm
import spharapy.spharatransform as st
import spharapy.spharafilter as sf
import spharapy.datasets as sd


# In[2]:


# import additional modules used in this tutorial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.offsetbox import AnchoredText
import numpy as np
from matplotlib.colors import LogNorm


# In[3]:


__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.11"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


# ## Import the spatial configuration of the EEG sensors and the SEP data

# In this tutorial we will apply a spatial SPHARA filter to SEP data of a single subject recorded with a 256 channel EEG system with equidistant layout. The data set is one of the example data sets contained in the SpharaPy toolbox.

# In[4]:


# loading the 256 channel EEG dataset from spharapy sample datasets
data_in = sd.load_eeg_256_channel_study()


# The dataset includes lists of vertices, triangles, and sensor labels, as well as EEG data from previously performed experiment addressing the cortical activation related to somatosensory-evoked potentials (SEP).

# In[5]:


print(data_in.keys())


# The triangulation of the EEG sensor setup consists of 256 vertices and 482 triangles. The EEG data consists of 256 channels and 369 time samples, 50 ms before to 130 ms after stimulation. The sampling frequency is 2048 Hz.

# In[6]:


vertlist = np.array(data_in['vertlist'])
trilist = np.array(data_in['trilist'])
eegdata = np.array(data_in['eegdata'])
print('vertices = ', vertlist.shape)
print('triangles = ', trilist.shape)
print('eegdata = ', eegdata.shape)


# ## Visualization of the EEG sensor setup and the measured data

# Visualization of triangulated sensor setup

# In[7]:


fig = plt.figure()
fig.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)
ax = fig.gca(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')
ax.set_title('The triangulated EEG sensor setup')
ax.view_init(elev=20., azim=80.)
ax.set_aspect('auto')

ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1], vertlist[:, 2],
                triangles=trilist, color='lightblue', edgecolor='black',
                linewidth=0.5, shade=True)
plt.show()
fig.savefig('sensor_setup.pdf')


# Visualization of the measured EEG data

# In[8]:


x = np.arange(-50, 130, 1/2.048)
figeeg = plt.figure()
axeeg = figeeg.gca()
axeeg.plot(x, eegdata[:, :].transpose())
axeeg.set_xlabel('t/ms')
axeeg.set_ylabel('V/µV')
axeeg.set_title('SEP data')
axeeg.set_ylim(-3.5, 3.5)
axeeg.set_xlim(-50, 130)
axeeg.grid(True)
plt.show()
figeeg.savefig('sep_butterfly.pdf')


# ## Create a SpharaPy TriMesh instance

# In the next step we create an instance of the class `spharapy.trimesh.TriMesh` using the list of vertices and triangles.

# In[9]:


# create an instance of the TriMesh class
mesh_eeg = tm.TriMesh(trilist, vertlist)


# ## SPHARA analysis of EEG data

# ### Create a SpharaPy SpharaTransform instance

# Afterwards we determine an instance of the class SpharaTransform, which is used to execute the transformation. For the determination of the SPHARA basis we use a Laplace-Beltrami operator, which is discretized by the FEM approach.

# In[10]:


sphara_transform_fem = st.SpharaTransform(mesh_eeg, 'fem')
basis_functions_fem, natural_frequencies_fem = sphara_transform_fem.basis()


# ### Visualization the basis functions

#  The first 15 spatially low-frequency SPHARA basis functions of the basis used for the transform are shown below, starting with DC at the top left.

# In[11]:


figsb1, axes1 = plt.subplots(nrows=3, ncols=3, figsize=(6.4, 6.4),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_fem[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    if i == 0:
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
    else:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_zticklabels([])
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('equal')
    ax.add_artist(AnchoredText('BF' + str(i), frameon=True, loc='upper center'))
    # ax.text2D(0, 0.8, 'local max')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.autoscale()
    trisurfplot.set_clim(-0.01, 0.01)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.95,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, 0))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.15, top=1.0, wspace=0.01, hspace=0.01)
plt.show()
figsb1.savefig('sphara_bf_lowfreq.pdf')


# ### SPHARA transform of the EEG data

# Subsequently we perform the SPHARA transformation of the EEG data. As a result, a butterfly plot of all channels of the EEG is compared to the visualization of the power contributions of the first 40 SPHARA basis functions. Only the first 40 out of 256 basis functions are used for the visualization, since the power contribution of the higher basis functions is very low.

# In[12]:


# perform the SPHARA transform
sphara_trans_eegdata = sphara_transform_fem.analysis(eegdata.transpose())


# In[13]:


# 60 low-frequency basis functions are displayed
ysel = 60
figsteeg, (axsteeg1, axsteeg2) = plt.subplots(nrows=2, figsize=(6.4, 4.8))

y = np.arange(0, ysel)
x = np.arange(-50, 130, 1/2.048)

axsteeg1.plot(x, eegdata[:, :].transpose())
axsteeg1.set_ylabel('V/µV')
axsteeg1.set_title('EEG data, 256 channels')
axsteeg1.set_ylim(-2.5, 2.5)
axsteeg1.set_xlim(-50, 130)
axsteeg1.grid(True)
axsteeg1.add_artist(AnchoredText("a", frameon=True, loc='upper left'))

pcm = axsteeg2.pcolormesh(x, y,
                          np.square(np.abs(sphara_trans_eegdata.transpose()
                                           [0:ysel, :])),
                         norm=LogNorm(vmin=0.1,vmax=25000),
                         linewidth=0.0,rasterized=True)
axsteeg2.set_xlabel('t/ms')
axsteeg2.set_ylabel('# BF')
axsteeg2.set_title('Power contribution of SPHARA basis functions')
axsteeg2.grid(True)
axsteeg2.add_artist(AnchoredText("b", frameon=True, loc='upper left'))
figsteeg.colorbar(pcm, ax=[axsteeg1, axsteeg2], shrink=0.45,
                  anchor=(0.85, 0.0), label='power / a.u.')

plt.subplots_adjust(left=0.1, right=0.85, bottom=0.1, top=0.95, hspace=0.35)
plt.show()
figsteeg.savefig('sep_butterfly_sphara_analysis.pdf')


# ## SPHARA filter using FEM discretisation

# ### Create a SpharaPy SpharaFilter instance

#  In the following step of the tutorial we determine an instance of the class SpharaFilter, which is used to execute the spatial filtering. For the determination of the SPHARA basis we use a Laplace-Beltrami operator, which is discretized by the FEM approach.

# In[14]:


sphara_filter_fem = sf.SpharaFilter(mesh_eeg, mode='fem',
                                    specification=20)


# ### SPHARA filtering of the EEG data

# In the next step we perform the SPHARA filtering of the EEG data. As a result, the butterfly plots of all channels of the EEG with and without filtering is compared. For the marked time samples also topo plots are provided.

# In[15]:


# perform the SPHARA filtering
sphara_filt_eegdata = sphara_filter_fem.filter(eegdata.transpose()).transpose()


# In[16]:


figsteeg, (axsteeg1, axsteeg2) = plt.subplots(nrows=2, figsize=(6.4, 6.4))

axsteeg1.axvline(13, color='red')
axsteeg1.axvline(19, color='blue')
axsteeg1.axvline(30, color='green')
axsteeg1.plot(x, eegdata[:, :].transpose())
axsteeg1.set_title('Unfiltered EEG data')
axsteeg1.set_ylabel('V/µV')
axsteeg1.set_xlabel('t/ms')
axsteeg1.set_ylim(-2.5, 2.5)
axsteeg1.set_xlim(-50, 130)
axsteeg1.grid(True)

axsteeg2.axvline(13, color='red')
axsteeg2.axvline(19, color='blue')
axsteeg2.axvline(30, color='green')
axsteeg2.plot(x, sphara_filt_eegdata[:, :].transpose())
axsteeg2.set_title('SPHARA low-pass filtered EEG data, 20 BF, fem')
axsteeg2.set_ylabel('V/µV')
axsteeg2.set_xlabel('t/ms')
axsteeg2.set_ylim(-2.5, 2.5)
axsteeg2.set_xlim(-50, 130)
axsteeg2.grid(True)

plt.subplots_adjust(left=0.1, right=0.95, bottom=0.1, top=0.95, hspace=0.3)
plt.show()
figsteeg.savefig('sep_butterfly_filt.pdf')


# In[17]:


time_pts = [129, 141, 164]
figsf1, axessf1 = plt.subplots(nrows=2, ncols=3, figsize=(6.4, 4.8),
                               subplot_kw={'projection': '3d'})

for i in range(2):
    for j in range(3):
        if i == 0:
            colorssf1 = np.mean(eegdata[trilist, time_pts[j]], axis=1)
        else:
            colorssf1 = np.mean(sphara_filt_eegdata[trilist, time_pts[j]],
                                axis=1)
        ax = axes1.flat[i]
        axessf1[i, j].set_xlabel('x')
        axessf1[i, j].set_ylabel('y')
        axessf1[i, j].set_zlabel('z')
        axessf1[i, j].view_init(elev=60., azim=80.)
        axessf1[i, j].set_aspect('equal')

        trisurfplot = axessf1[i, j].plot_trisurf(vertlist[:, 0],
                                                 vertlist[:, 1],
                                                 vertlist[:, 2],
                                                 triangles=trilist,
                                                 cmap=plt.cm.bwr,
                                                 edgecolor='white',
                                                 linewidth=0.)
        trisurfplot.set_array(colorssf1)
        trisurfplot.set_clim(-2., 2)

cbar = figsb1.colorbar(trisurfplot, ax=axessf1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, 0.0))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.225, top=0.95,
                    wspace=0.0, hspace=0.0)
plt.show()
figsf1.savefig('sep_topo_filt.pdf')


# ### Application of the the spatial SPHARA filter to data with artificial noise

# In a final step the EEG data are disturbed by white noise with different noise levels (3dB, 0dB and -3dB). A spatial low-pass SPHARA filter with 20 basis functions is applied to these data.

# In[18]:


# vector with noise levels in dB
db_val_vec = [3, 0, -3]

# compute the power of the SEP data
power_sep = np.sum(np.square(np.absolute(eegdata))) / eegdata.size

# compute a vector with standard deviations of the noise in relation
# to signal power for the given noise levels
noise_sd_vec = list(map(lambda db_val:
                    np.sqrt(power_sep / (10 ** (db_val / 10))),
                        db_val_vec))

# add the noise to the EEG data
eegdata_noise = list(map(lambda noise_sd:
                         eegdata + np.random.normal(0, noise_sd, [256, 369]),
                         noise_sd_vec))

# filter the EEG data containing the artificial noise
eegdata_noise_filt = list(map(lambda eeg_noise:
                              (sphara_filter_fem.filter(eeg_noise.transpose()).
                               transpose()),
                              eegdata_noise))


#  The results of the filtering are shown below.

# In[19]:


figfilt, axesfilt = plt.subplots(nrows=4, ncols=2, figsize=(6.4, 8.0))

axesfilt[0, 0].plot(x, eegdata[:, :].transpose())
axesfilt[0, 0].set_title('EEG')
axesfilt[0, 0].set_ylabel('V/µV')
# axesfilt[0, 0].set_xlabel('t/ms')
axesfilt[0, 0].set_ylim(-2.5, 2.5)
axesfilt[0, 0].set_xlim(-50, 130)
axesfilt[0, 0].grid(True)

axesfilt[0, 1].plot(x, sphara_filt_eegdata[:, :].transpose())
axesfilt[0, 1].set_title('SPHARA low-pass filtered EEG')
# axesfilt[0, 1].set_ylabel('V/µV')
# axesfilt[0, 1].set_xlabel('t/ms')
axesfilt[0, 1].set_ylim(-2.5, 2.5)
axesfilt[0, 1].set_xlim(-50, 130)
axesfilt[0, 1].grid(True)

for i in range(3):
    axesfilt[i + 1, 0].plot(x, eegdata_noise[i].transpose())
    axesfilt[i + 1, 0].set_title('EEG + noise, SNR ' +
                                 str(db_val_vec[i]) + 'dB')
    axesfilt[i + 1, 0].set_ylabel('V/µV')
    if i == 2:
        axesfilt[i + 1, 0].set_xlabel('t/ms')
    axesfilt[i + 1, 0].set_ylim(-2.5, 2.5)
    axesfilt[i + 1, 0].set_xlim(-50, 130)
    axesfilt[i + 1, 0].grid(True)

    axesfilt[i + 1, 1].plot(x, eegdata_noise_filt[i].transpose())
    axesfilt[i + 1, 1].set_title('EEG + noise, SNR ' +
                                 str(db_val_vec[i]) + 'dB, SPHARA filtered')
    # axesfilt[i + 1, 1].set_ylabel('V/µV')
    if i == 2:
        axesfilt[i + 1, 1].set_xlabel('t/ms')
    axesfilt[i + 1, 1].set_ylim(-2.5, 2.5)
    axesfilt[i + 1, 1].set_xlim(-50, 130)
    axesfilt[i + 1, 1].grid(True)

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.0, top=1.0, hspace=0.35,
                   wspace=0.15)
plt.show()
figfilt.savefig('sep_butterfly_noise_sim_filt.pdf')




# -*- coding: utf-8 -*-

"""
The :mod:`spharapy.datasets`: module includes utilities to provide
sample datasets.
"""

from .base import load_minimal_triangular_mesh
from .base import load_simple_triangular_mesh
from .base import load_eeg_256_channel_study


__author__ = "Uwe Graichen"
__copyright__ = "Copyright 2018-2019, Uwe Graichen"
__credits__ = ["Uwe Graichen"]
__license__ = "BSD-3-Clause"
__version__ = "1.0.13"
__maintainer__ = "Uwe Graichen"
__email__ = "uwe.graichen@tu-ilmenau.de"
__status__ = "Release"


__all__ = ['load_minimal_triangular_mesh',
           'load_simple_triangular_mesh',
           'load_eeg_256_channel_study']



# -*- coding: utf-8 -*-
r""".. _quick_start:


Quick start with SpharaPy
=========================

.. topic:: Section contents

   In this tutorial, we briefly introduce the vocabulary used in
   spatial harmonic analysis (SPHARA) and we give a simple learning
   example to SpharaPy.


SPHARA -- The problem setting
-----------------------------

Fourier analysis is one of the standard tools in digital signal and
image processing. In ordinary digital image data, the pixels are
arranged in a Cartesian or rectangular grid. Performing the Fourier
transform, the image data :math:`x[m,n]` is compared (using a scalar
product) with a two-dimensional Fourier basis :math:`f[k,l] =
\mathrm{e}^{-2\pi \mathrm{i} \cdot \left(\frac{mk}{M} + \frac{nl}{N}
\right) }`. In Fourier transform on a Cartesian grid, the Fourier
basis used is usually inherently given in the transformation rule

.. math::

   X[k,l] = \sum_{m=0}^{M-1} \sum_{n=0}^{N-1} x[m,n] \cdot
   \mathrm{e}^{-2\pi \mathrm{i} \cdot \left(\frac{mk}{M} +
   \frac{nl}{N} \right) }\,.

A Fourier basis is a solution to Laplace's eigenvalue problem (related
to the Helmholtz equation)

.. math::

   L \vec{x} = \lambda \vec{x}\,,\qquad\qquad(1)

with the discrete :term:`Laplace-Beltrami operator` in matrix notation
:math:`L \in \mathbb{R}^{M \times N}`, the eigenvectors
:math:`\vec{x}` containing the harmonic functions and the eigenvalues
:math:`\lambda` the natural frequencies. 



An arbitrary arrangement of sample points on a surface in
three-dimensional space can be described by means of a
:term:`triangular mesh`. A spatial harmonic basis (**SPHARA basis**)
is a solution of a Laplace eigenvalue problem for the given triangle
mesh can be obtained by discretizing a Laplace-Beltrami operator for
the mesh and solving the Laplace eigenvalue problem in equation
(1). SpharaPy provides classes and functions to support these tasks:

- managing triangular meshes describing the spatial arrangement of
  the sample points,
- determining the Laplace-Beltrami operator of these meshes,
- computing a basis for spatial Fourier analysis of data defined
  on the triangular mesh, and
- performing the SPHARA transform and filtering.

"""

######################################################################
# The SpharaPy package
# --------------------
# The SpharaPy package consists of five modules
# :mod:`spharapy.trimesh`, :mod:`spharapy.spharabasis`,
# :mod:`spharapy.spharatransform`, :mod:`spharapy.spharafilter` and
# :mod:`spharapy.datasets`. In the following we use three of the five
# SpharaPy modules to briefly show how a SPHARA basis can be calculated
# for given spatial sample points. The :mod:`spharapy.trimesh` module
# contains the TriMesh class, which can be used to specify the
# configuration of the spatial sample points. The SPHARA basis
# functions can be determined using the :mod:`spharapy.spharabasis`
# module, employing different discretizations. The
# :mod:`spharapy.datasets` module is an interface to the example data
# sets provided with the SpharaPy package.

# Code source: Uwe Graichen
# License: BSD 3 clause

# import modules from spharapy package
import spharapy.trimesh as tm
import spharapy.spharabasis as sb
import spharapy.datasets as sd

# import additional modules used in this tutorial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np

######################################################################
# Specification of the spatial configuration of the sample points
# ---------------------------------------------------------------
#
# To illustrate some basic functionality of the SpharaPy package, we
# load a simple triangle mesh from the example data sets.

# loading the simple mesh from spharapy sample datasets
mesh_in = sd.load_simple_triangular_mesh()

######################################################################
# The imported mesh is defined by a **list of triangles** and a **list of
# vertices**. The data are stored in a dictionary with the two keys
# 'vertlist' and 'trilist'

print(mesh_in.keys())

######################################################################
# The simple, triangulated surface consists of 131 vertices and 232
# triangles and is the triangulation of a hemisphere of an unit ball.

vertlist = np.array(mesh_in['vertlist'])
trilist = np.array(mesh_in['trilist'])
print('vertices = ', vertlist.shape)
print('triangles = ', trilist.shape)

######################################################################

fig = plt.figure()
fig.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)
ax = fig.gca(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')
ax.view_init(elev=60., azim=45.)
ax.set_aspect('auto')
ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1], vertlist[:, 2],
                triangles=trilist, color='lightblue', edgecolor='black',
                linewidth=1)


######################################################################
#
# Determining the Laplace-Beltrami Operator
# -----------------------------------------
#
# In a further step, an instance of the class
# :class:`spharapy.trimesh.TriMesh` is created from the lists of
# vertices and triangles. The class :class:`spharapy.trimesh.TriMesh`
# provides a number of methods to determine certain properties of the
# triangle mesh required to generate the SPHARA basis.

# print all implemented methods of the TriMesh class
print([func for func in dir(tm.TriMesh) if not func.startswith('__')])

######################################################################

# create an instance of the TriMesh class
simple_mesh = tm.TriMesh(trilist, vertlist)

######################################################################
# For the simple triangle mesh an instance of the class SpharaBasis is
# created and the finite element discretization ('fem') is used. The
# complete set of SPHARA basis functions and the natural frequencies
# associated with the basis functions are determined.

sphara_basis = sb.SpharaBasis(simple_mesh, 'fem')
basis_functions, natural_frequencies = sphara_basis.basis()

######################################################################
# The set of SPHARA basis functions can be used for spatial Fourier
# analysis of the spatially irregularly sampled data.
#
# The first 15 spatially low-frequency SPHARA basis functions are
# shown below, starting with DC at the top left.


# sphinx_gallery_thumbnail_number = 2
figsb1, axes1 = plt.subplots(nrows=5, ncols=3, figsize=(8, 12),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=70., azim=15.)
    ax.set_aspect('auto')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.set_clim(-1, 1)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.75,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, -4.0))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.08, top=1.0)
plt.show()



# -*- coding: utf-8 -*-
r""".. _sphara_filtering_eeg:


Spatial SPHARA filtering of EEG data
====================================

.. topic:: Section contents

   In this tutorial we show how to use the SPHARA basis functions to
   design a spatial low pass filter for application to EEG data. The
   FEM discretization of the Laplace-Beltrami operator is used to
   calculate the SPHARA basic functions that are used for the SPHARA
   low pass filter. The applicability of the filter is shown using an
   EEG data set that is disturbed by white noise in different noise
   levels.


Introduction
------------

The human head as a volume conductor exhibits spatial low-pass filter
properties. For this reason, the potential distribution of the EEG on
the scalp surface can be represented by a few low-frequency SPHARA
basis functions, compare :ref:`sphara_analysis_eeg`. In contrast,
single channel dropouts and spatially uncorrelated sensor noise
exhibit an almost equally distributed spatial SPHARA spectrum. This
fact can be exploited for the design of a spatial filter for the
suppression of uncorrelated sensor noise.

"""

######################################################################
# At the beginning we import three modules of the SpharaPy package as
# well as several other packages and single functions from
# packages.

# Code source: Uwe Graichen
# License: BSD 3 clause

# import modules from spharapy package
import spharapy.trimesh as tm
import spharapy.spharafilter as sf
import spharapy.datasets as sd

# import additional modules used in this tutorial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np


######################################################################
# Import the spatial configuration of the EEG sensors and the SEP data
# --------------------------------------------------------------------
# In this tutorial we will apply a spatial SPHARA filter to SEP data
# of a single subject recorded with a 256 channel EEG system with
# equidistant layout. The data set is one of the example data sets
# contained in the SpharaPy toolbox.

# loading the 256 channel EEG dataset from spharapy sample datasets
mesh_in = sd.load_eeg_256_channel_study()

######################################################################
# The dataset includes lists of vertices, triangles, and sensor
# labels, as well as EEG data from previously performed experiment
# addressing the cortical activation related to somatosensory-evoked
# potentials (SEP).

print(mesh_in.keys())

######################################################################
# The triangulation of the EEG sensor setup consists of 256 vertices
# and 480 triangles. The EEG data consists of 256 channels and 369
# time samples, 50 ms before to 130 ms after stimulation. The sampling
# frequency is 2048 Hz.

vertlist = np.array(mesh_in['vertlist'])
trilist = np.array(mesh_in['trilist'])
eegdata = np.array(mesh_in['eegdata'])
print('vertices = ', vertlist.shape)
print('triangles = ', trilist.shape)
print('eegdata = ', eegdata.shape)

######################################################################

fig = plt.figure()
fig.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)
ax = fig.gca(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')
ax.set_title('The triangulated EEG sensor setup')
ax.view_init(elev=20., azim=80.)
ax.set_aspect('auto')
ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1], vertlist[:, 2],
                triangles=trilist, color='lightblue', edgecolor='black',
                linewidth=0.5, shade=True)
plt.show()

######################################################################

x = np.arange(-50, 130, 1/2.048)
figeeg = plt.figure()
axeeg = figeeg.gca()
axeeg.plot(x, eegdata[:, :].transpose())
axeeg.set_xlabel('t/ms')
axeeg.set_ylabel('V/µV')
axeeg.set_title('SEP data')
axeeg.set_ylim(-3.5, 3.5)
axeeg.set_xlim(-50, 130)
axeeg.grid(True)
plt.show()


######################################################################
# Create a SpharaPy TriMesh instance
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# In the next step we create an instance of the class
# :class:`spharapy.trimesh.TriMesh` from the list of vertices and
# triangles.

# create an instance of the TriMesh class
mesh_eeg = tm.TriMesh(trilist, vertlist)


######################################################################
# SPHARA filter using FEM discretisation
# --------------------------------------
# Create a SpharaPy SpharaFilter instance
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In the following step of the tutorial we determine an instance of
# the class SpharaFilter, which is used to execute the spatial
# filtering. For the determination of the SPHARA basis we use a
# Laplace-Beltrami operator, which is discretized by the FEM approach.

sphara_filter_fem = sf.SpharaFilter(mesh_eeg, mode='fem',
                                    specification=20)
basis_functions_fem, natural_frequencies_fem = sphara_filter_fem.basis()

######################################################################
# Visualization the basis functions
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# The first 15 spatially low-frequency SPHARA basis functions are
# shown below, starting with DC at the top left.
#

figsb1, axes1 = plt.subplots(nrows=5, ncols=3, figsize=(8, 12),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_fem[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('auto')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.set_clim(-0.01, 0.01)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, -4.5))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.08, top=1.0)
plt.show()


######################################################################
# SPHARA filtering of the EEG data
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In the next step we perform the SPHARA filtering of the EEG
# data. As a result, the butterfly plots of all channels of the EEG
# with and without filtering is compared. For the marked time samples
# also topo plots are provided.

# perform the SPHARA filtering
sphara_filt_eegdata = sphara_filter_fem.filter(eegdata.transpose()).transpose()

figsteeg, (axsteeg1, axsteeg2) = plt.subplots(nrows=2, figsize=(8, 6.5))

axsteeg1.axvline(13, color='red')
axsteeg1.axvline(19, color='blue')
axsteeg1.axvline(30, color='green')
axsteeg1.plot(x, eegdata[:, :].transpose())
axsteeg1.set_title('Unfiltered EEG data')
axsteeg1.set_ylabel('V/µV')
axsteeg1.set_xlabel('t/ms')
axsteeg1.set_ylim(-2.5, 2.5)
axsteeg1.set_xlim(-50, 130)
axsteeg1.grid(True)

axsteeg2.axvline(13, color='red')
axsteeg2.axvline(19, color='blue')
axsteeg2.axvline(30, color='green')
axsteeg2.plot(x, sphara_filt_eegdata[:, :].transpose())
axsteeg2.set_title('SPHARA low-pass filtered EEG data, 20 BF, fem')
axsteeg2.set_ylabel('V/µV')
axsteeg2.set_xlabel('t/ms')
axsteeg2.set_ylim(-2.5, 2.5)
axsteeg2.set_xlim(-50, 130)
axsteeg2.grid(True)

plt.subplots_adjust(left=0.1, right=0.95, bottom=0.1, top=0.95, hspace=0.3)
plt.show()

######################################################################

time_pts = [129, 141, 164]
figsf1, axessf1 = plt.subplots(nrows=2, ncols=3, figsize=(8, 5),
                               subplot_kw={'projection': '3d'})

for i in range(2):
    for j in range(3):
        if i == 0:
            colorssf1 = np.mean(eegdata[trilist, time_pts[j]], axis=1)
        else:
            colorssf1 = np.mean(sphara_filt_eegdata[trilist, time_pts[j]],
                                axis=1)
        ax = axes1.flat[i]
        axessf1[i, j].set_xlabel('x')
        axessf1[i, j].set_ylabel('y')
        axessf1[i, j].set_zlabel('z')
        axessf1[i, j].view_init(elev=60., azim=80.)
        axessf1[i, j].set_aspect('auto')

        trisurfplot = axessf1[i, j].plot_trisurf(vertlist[:, 0],
                                                 vertlist[:, 1],
                                                 vertlist[:, 2],
                                                 triangles=trilist,
                                                 cmap=plt.cm.bwr,
                                                 edgecolor='white',
                                                 linewidth=0.)
        trisurfplot.set_array(colorssf1)
        trisurfplot.set_clim(-2., 2)

cbar = figsb1.colorbar(trisurfplot, ax=axessf1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, 0))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.2, top=1.0)
plt.show()


######################################################################
# Application of the the spatial SPHARA filter to data with artificial noise
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In a final step the EEG data are disturbed by white noise with
# different noise levels (3dB, 0dB and -3dB). A spatial low-pass
# SPHARA filter with 20 basis functions is applied to these data. The
# results of the filtering are shown below.

# vector with noise levels in dB
db_val_vec = [3, 0, -3]

# compute the power of the SEP data
power_sep = np.sum(np.square(np.absolute(eegdata))) / eegdata.size

# compute a vector with standard deviations of the noise in relation
# to signal power for the given noise levels
noise_sd_vec = list(map(lambda db_val:
                    np.sqrt(power_sep / (10 ** (db_val / 10))),
                        db_val_vec))

# add the noise to the EEG data
eegdata_noise = list(map(lambda noise_sd:
                         eegdata + np.random.normal(0, noise_sd, [256, 369]),
                         noise_sd_vec))

# filter the EEG data containing the artificial noise
eegdata_noise_filt = list(map(lambda eeg_noise:
                              (sphara_filter_fem.filter(eeg_noise.transpose()).
                               transpose()),
                              eegdata_noise))

######################################################################

figfilt, axesfilt = plt.subplots(nrows=4, ncols=2, figsize=(8, 10.5))

axesfilt[0, 0].plot(x, eegdata[:, :].transpose())
axesfilt[0, 0].set_title('EEG data')
axesfilt[0, 0].set_ylabel('V/µV')
axesfilt[0, 0].set_xlabel('t/ms')
axesfilt[0, 0].set_ylim(-2.5, 2.5)
axesfilt[0, 0].set_xlim(-50, 130)
axesfilt[0, 0].grid(True)

axesfilt[0, 1].plot(x, sphara_filt_eegdata[:, :].transpose())
axesfilt[0, 1].set_title('SPHARA low-pass filtered EEG data')
axesfilt[0, 1].set_ylabel('V/µV')
axesfilt[0, 1].set_xlabel('t/ms')
axesfilt[0, 1].set_ylim(-2.5, 2.5)
axesfilt[0, 1].set_xlim(-50, 130)
axesfilt[0, 1].grid(True)

for i in range(3):
    axesfilt[i + 1, 0].plot(x, eegdata_noise[i].transpose())
    axesfilt[i + 1, 0].set_title('EEG data + noise, SNR ' +
                                 str(db_val_vec[i]) + 'dB')
    axesfilt[i + 1, 0].set_ylabel('V/µV')
    axesfilt[i + 1, 0].set_xlabel('t/ms')
    axesfilt[i + 1, 0].set_ylim(-2.5, 2.5)
    axesfilt[i + 1, 0].set_xlim(-50, 130)
    axesfilt[i + 1, 0].grid(True)

    axesfilt[i + 1, 1].plot(x, eegdata_noise_filt[i].transpose())
    axesfilt[i + 1, 1].set_title('EEG data + noise, SNR ' +
                                 str(db_val_vec[i]) + 'dB, SPHARA filtered')
    axesfilt[i + 1, 1].set_ylabel('V/µV')
    axesfilt[i + 1, 1].set_xlabel('t/ms')
    axesfilt[i + 1, 1].set_ylim(-2.5, 2.5)
    axesfilt[i + 1, 1].set_xlim(-50, 130)
    axesfilt[i + 1, 1].grid(True)

plt.subplots_adjust(left=0.07, right=0.97, bottom=0.05, top=0.95, hspace=0.45)
plt.show()
# sphinx_gallery_thumbnail_number = 6



# -*- coding: utf-8 -*-
r""".. _sphara_analysis_eeg:


Spatial SPHARA analysis of EEG data
===================================

.. topic:: Section contents

   This tutorial shows exemplarily the spatial SPHARA analysis of
   256-channel EEG data. The FEM discretization of the
   Laplace-Beltrami operator is employed to calculate the SPHARA basic
   functions that are used for the SPHARA decomposition.


Introduction
------------

As explained in :ref:`introduction` and in the tutorial
:ref:`sphara_basis_eeg` a spatial Fourier basis for a arbitrary sensor
setup can be determined as solution of the Laplace's eigenvalue
problem dicretized for the considered sensor setup

.. math::

   L \vec{x} = \lambda \vec{x}\,,

with the discrete :term:`Laplace-Beltrami operator` in matrix notation
:math:`L \in \mathbb{R}^{M \times N}`, the eigenvectors
:math:`\vec{x}` containing the harmonic functions and the eigenvalues
:math:`\lambda` the natural frequencies.

The spatial Fourier basis determined in this way can be used for the
spatial Fourier analysis of data recorded with the considered sensor
setup.

For the anaylsis of discrete data defined on the vertices of the
triangular mesh - the SPHARA transform - the inner product is used
(transformation from spatial domain to spatial frequency domain). For
an analysis using eigenvectors computed by the FEM approach, the inner
product that assures the :math:`B`-orthogonality needs to be applied.

For the reverse transformation, the discrete data are synthesized
using the linear combination of the SPHARA coefficients and the
corresponding SPHARA basis functions. More detailed information can be
found in the section :ref:`sphara_analysis_synthesis` and in
:cite:`graichen15`.

"""

######################################################################
# At the beginning we import three modules of the SpharaPy package as
# well as several other packages and single functions of packages.

# Code source: Uwe Graichen
# License: BSD 3 clause

# import modules from spharapy package
import spharapy.trimesh as tm
import spharapy.spharatransform as st
import spharapy.datasets as sd

# import additional modules used in this tutorial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np


######################################################################
# Import the spatial configuration of the EEG sensors and the SEP data
# --------------------------------------------------------------------
# In this tutorial we will apply the SPHARA analysis to SEP data of a
# single subject recorded with a 256 channel EEG system with
# equidistant layout. The data set is one of the example data sets
# contained in the SpharaPy toolbox.

# loading the 256 channel EEG dataset from spharapy sample datasets
mesh_in = sd.load_eeg_256_channel_study()

######################################################################
# The dataset includes lists of vertices, triangles, and sensor
# labels, as well as EEG data from previously performed experiment
# addressing the cortical activation related to somatosensory-evoked
# potentials (SEP).

print(mesh_in.keys())

######################################################################
# The triangulation of the EEG sensor setup consists of 256 vertices
# and 480 triangles. The EEG data consists of 256 channels and 369
# time samples, 50 ms before to 130 ms after stimulation. The sampling
# frequency is 2048 Hz.

vertlist = np.array(mesh_in['vertlist'])
trilist = np.array(mesh_in['trilist'])
eegdata = np.array(mesh_in['eegdata'])
print('vertices = ', vertlist.shape)
print('triangles = ', trilist.shape)
print('eegdata = ', eegdata.shape)

######################################################################

fig = plt.figure()
fig.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)
ax = fig.gca(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')
ax.set_title('The triangulated EEG sensor setup')
ax.view_init(elev=20., azim=80.)
ax.set_aspect('auto')
ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1], vertlist[:, 2],
                triangles=trilist, color='lightblue', edgecolor='black',
                linewidth=0.5, shade=True)
plt.show()

######################################################################

x = np.arange(-50, 130, 1/2.048)
figeeg = plt.figure()
axeeg = figeeg.gca()
axeeg.plot(x, eegdata[:, :].transpose())
axeeg.set_xlabel('t/ms')
axeeg.set_ylabel('V/µV')
axeeg.set_title('SEP data')
axeeg.set_ylim(-3.5, 3.5)
axeeg.set_xlim(-50, 130)
axeeg.grid(True)
plt.show()


######################################################################
# Create a SpharaPy TriMesh instance
# ----------------------------------
#
# In the next step we create an instance of the class
# :class:`spharapy.trimesh.TriMesh` from the list of vertices and
# triangles.

# create an instance of the TriMesh class
mesh_eeg = tm.TriMesh(trilist, vertlist)


######################################################################
# SPHARA transform using FEM discretisation
# -----------------------------------------
# Create a SpharaPy SpharaTransform instance
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In the next step of the tutorial we determine an instance of the
# class SpharaTransform, which is used to execute the
# transformation. For the determination of the SPHARA basis we use a
# Laplace-Beltrami operator, which is discretized by the FEM
# approach.

sphara_transform_fem = st.SpharaTransform(mesh_eeg, 'fem')
basis_functions_fem, natural_frequencies_fem = sphara_transform_fem.basis()

######################################################################
# Visualization the basis functions
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# The first 15 spatially low-frequency SPHARA basis functions of the
# basis used for the transform are shown below, starting with DC
# at the top left.

figsb1, axes1 = plt.subplots(nrows=5, ncols=3, figsize=(8, 12),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_fem[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('auto')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.autoscale()
    trisurfplot.set_clim(-0.01, 0.01)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, -4.5))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.08, top=1.0)
plt.show()


######################################################################
# SPHARA transform of the EEG data
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In the final step we perform the SPHARA transformation of the EEG
# data. As a result, a butterfly plot of all channels of the EEG is
# compared to the visualization of the power contributions of the
# first 40 SPHARA basis functions. Only the first 40 out of 256 basis
# functions are used for the visualization, since the power
# contribution of the higher basis functions is very low.

# perform the SPHARA transform
sphara_trans_eegdata = sphara_transform_fem.analysis(eegdata.transpose())

# 40 low-frequency basis functions are displayed
ysel = 40
figsteeg, (axsteeg1, axsteeg2) = plt.subplots(nrows=2)

y = np.arange(0, ysel)
x = np.arange(-50, 130, 1/2.048)

axsteeg1.plot(x, eegdata[:, :].transpose())
axsteeg1.set_ylabel('V/µV')
axsteeg1.set_title('EEG data, 256 channels')
axsteeg1.set_ylim(-2.5, 2.5)
axsteeg1.set_xlim(-50, 130)
axsteeg1.grid(True)

pcm = axsteeg2.pcolormesh(x, y,
                          np.square(np.abs(sphara_trans_eegdata.transpose()
                                           [0:ysel, :])))
axsteeg2.set_xlabel('t/ms')
axsteeg2.set_ylabel('# BF')
axsteeg2.set_title('Power contribution of SPHARA basis functions')
axsteeg2.grid(True)
figsteeg.colorbar(pcm, ax=[axsteeg1, axsteeg2], shrink=0.45,
                  anchor=(0.85, 0.0), label='power / a.u.')

plt.subplots_adjust(left=0.1, right=0.85, bottom=0.1, top=0.95, hspace=0.35)
plt.show()
# sphinx_gallery_thumbnail_number = 4



# -*- coding: utf-8 -*-
r""".. _sphara_basis_eeg:


Determination of the SPHARA basis functions for an EEG sensor setup
===================================================================

.. topic:: Section contents

   This tutorial introduces the steps necessary to determine a
   generalized spatial Fourier basis for an :term:`EEG` sensor setup
   using SpharaPy. The special properties of the different
   discretization approaches of the Laplace-Beltrami operator will be
   discussed.


Introduction
------------

A Fourier basis is a solution to Laplace's eigenvalue problem

.. math::

   L \vec{x} = \lambda \vec{x}\,,\qquad\qquad(1)

with the discrete :term:`Laplace-Beltrami operator` in matrix notation
:math:`L \in \mathbb{R}^{M \times N}`, the eigenvectors
:math:`\vec{x}` containing the harmonic functions and the eigenvalues
:math:`\lambda` the natural frequencies.

By solving a Laplace eigenvalue problem, it is also possible to
determine a basis for a spatial Fourier analysis. Often in practical
applications, a measured quantity to be subjected to a Fourier
analysis is only known at spatially discrete sampling points (the
sensor positions). An arbitrary arrangement of sample points on a
surface in three-dimensional space can be described by means of a
:term:`triangular mesh`. In the case of an :term:`EEG` system, the
sample positions (the vertices of the triangular mesh) are the
locations of the sensors arranged on the head surface. A SPHARA basis
is a solution of a Laplace eigenvalue problem for the given triangle
mesh, that can be obtained by discretizing a Laplace-Beltrami operator
for the mesh and solving the Laplace eigenvalue problem in equation
(1). The SpharaPy package provides three methods for the
discretization of the Laplace-Beltrami operator; unit weighting of the
edges and weighting with the inverse of the Euclidean distance of the
edges, and a FEM approach. For more detailed information please refer
to :ref:`introduction` and :ref:`eigensystems_of_lb_operators`.

"""

######################################################################
# At the beginning we import three modules of the SpharaPy package as
# well as several other packages and single functions of packages.

# Code source: Uwe Graichen
# License: BSD 3 clause

# import modules from spharapy package
import spharapy.trimesh as tm
import spharapy.spharabasis as sb
import spharapy.datasets as sd

# import additional modules used in this tutorial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np


######################################################################
# Specification of the spatial configuration of the EEG sensors
# -------------------------------------------------------------
#
# Import information about EEG sensor setup of the sample data set
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In this tutorial we will determine a SPHARA basis for a 256 channel
# :term:`EEG` system with equidistant layout. The data set is one of
# the example data sets contained in the SpharaPy toolbox, see
# :mod:`spharapy.datasets` and
# :func:`spharapy.datasets.load_eeg_256_channel_study`.

# loading the 256 channel EEG dataset from spharapy sample datasets
mesh_in = sd.load_eeg_256_channel_study()

######################################################################
# The dataset includes lists of vertices, triangles, and sensor
# labels, as well as :term:`EEG` data from previously performed
# experiment addressing the cortical activation related to
# somatosensory-evoked potentials (SEP).

print(mesh_in.keys())

######################################################################
# The triangulation of the :term:`EEG` sensor setup consists of 256
# vertices and 480 triangles.

vertlist = np.array(mesh_in['vertlist'])
trilist = np.array(mesh_in['trilist'])
print('vertices = ', vertlist.shape)
print('triangles = ', trilist.shape)

######################################################################

fig = plt.figure()
fig.subplots_adjust(left=0.02, right=0.98, top=0.98, bottom=0.02)
ax = fig.gca(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')
ax.set_title('The triangulated EEG sensor setup')
ax.view_init(elev=20., azim=80.)
ax.set_aspect('auto')
ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1], vertlist[:, 2],
                triangles=trilist, color='lightblue', edgecolor='black',
                linewidth=0.5, shade=True)
plt.show()

######################################################################
# Create a SpharaPy TriMesh instance
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# In the next step we create an instance of the class
# :class:`spharapy.trimesh.TriMesh` from the list of vertices and
# triangles.

# create an instance of the TriMesh class
mesh_eeg = tm.TriMesh(trilist, vertlist)

######################################################################
# The class :class:`spharapy.trimesh.TriMesh` provides a
# number of methods to determine certain properties of the triangle
# mesh required to generate the SPHARA basis, listed below:

# print all implemented methods of the TriMesh class
print([func for func in dir(tm.TriMesh) if not func.startswith('__')])

######################################################################
# Determining SPHARA bases using different discretisation approaches
# ------------------------------------------------------------------
# Computing the basis functions
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# In the final step of the tutorial we will calculate SPHARA bases for
# the given :term:`EEG` sensor setup. For this we create three
# instances of the class :class:`spharapy.spharabasis.SpharaBasis`. We
# use the three discretization approaches implemented in this class
# for the Laplace-Beltrami operator: unit weighting ('unit') and
# inverse Euclidean weigthing ('inv_euclidean') of the edges of the
# triangular mesh as well as the FEM discretization ('fem')

# 'unit' discretization
sphara_basis_unit = sb.SpharaBasis(mesh_eeg, 'unit')
basis_functions_unit, natural_frequencies_unit = sphara_basis_unit.basis()

# 'inv_euclidean' discretization
sphara_basis_ie = sb.SpharaBasis(mesh_eeg, 'inv_euclidean')
basis_functions_ie, natural_frequencies_ie = sphara_basis_ie.basis()

# 'fem' discretization
sphara_basis_fem = sb.SpharaBasis(mesh_eeg, 'fem')
basis_functions_fem, natural_frequencies_fem = sphara_basis_fem.basis()

######################################################################
# Visualization the basis functions
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# The first 15 spatially low-frequency SPHARA basis functions are
# shown below, starting with DC at the top left.
#
# SPHARA basis using the discretization approache 'unit'
# """"""""""""""""""""""""""""""""""""""""""""""""""""""

# sphinx_gallery_thumbnail_number = 2
figsb1, axes1 = plt.subplots(nrows=5, ncols=3, figsize=(8, 12),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_unit[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('auto')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.set_clim(-0.15, 0.15)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, -4.5))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.08, top=1.0)
plt.show()

######################################################################
# SPHARA basis using the discretization approache 'inv_euclidean'
# """""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

figsb1, axes1 = plt.subplots(nrows=5, ncols=3, figsize=(8, 12),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_ie[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('auto')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.set_clim(-0.18, 0.18)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, -4.5))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.08, top=1.0)
plt.show()

######################################################################
# SPHARA basis using the discretization approache 'fem'
# """""""""""""""""""""""""""""""""""""""""""""""""""""

figsb1, axes1 = plt.subplots(nrows=5, ncols=3, figsize=(8, 12),
                             subplot_kw={'projection': '3d'})
for i in range(np.size(axes1)):
    colors = np.mean(basis_functions_fem[trilist, i + 0], axis=1)
    ax = axes1.flat[i]
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=60., azim=80.)
    ax.set_aspect('auto')
    trisurfplot = ax.plot_trisurf(vertlist[:, 0], vertlist[:, 1],
                                  vertlist[:, 2], triangles=trilist,
                                  cmap=plt.cm.bwr,
                                  edgecolor='white', linewidth=0.)
    trisurfplot.set_array(colors)
    trisurfplot.set_clim(-0.01, 0.01)

cbar = figsb1.colorbar(trisurfplot, ax=axes1.ravel().tolist(), shrink=0.85,
                       orientation='horizontal', fraction=0.05, pad=0.05,
                       anchor=(0.5, -4.5))

plt.subplots_adjust(left=0.0, right=1.0, bottom=0.08, top=1.0)
plt.show()



'''
Created on 04.07.2018

@author: graichen
'''
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


with open('README.rst') as f:
    readme = f.read()

setup(
    name='SpharaPy',
    version='1.0.15',
    description='SPHARA Implementation in Python',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Uwe Graichen',
    author_email='uwe.graichen@tu-ilmenau.de',
    url=' ',
    license='BSD-3-Clause',
    packages=find_packages(exclude=('tests', 'docs')),
    package_data={'spharapy': ['datasets/data/*.csv',
                               'datasets/descr/*.rst',
                               'examples/*.py',
                               'examples/*.ipynb',
                               'examples/images/*']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
    ],

    project_urls={  # Optional
        'Documentation': 'https://spharapy.readthedocs.io/en/latest/',
    },
)
